#!/usr/bin/env python

######################################################
# ratRunner.py
# ---------
# Author: Matt Mottram
#         <m.mottram@sussex.ac.uk>
#
# Description:
#    Runs user RAT jobs as a Ganga submitted job.
#
# Ships with all RATUser jobs.
# Allows users to install rat snapshots to the temporary
# job directory on the submission backend.  Can use a token
# or a tar.gz file that ships with the job to install.
# Script requires knowledge of the base version of RAT
# that the snapshot derives from (and creates an environment
# file accordingly).  Base version must be installed on the
# backend in a named location.
#
# Revision History:
#  - 2012/11/03: M. Mottram: first revision with proper documentation!
#  - 2014/05/19: M. Mottram: moved common methods to job_tools
#
######################################################

import os
import sys
import optparse
import glob
import socket
import json
import job_tools


#################################
# Deprecated functions
#################################
def download_local(token, run_version):
    '''Download copy of RAT.
    Should use method to ship with job instead.
    '''
    install_path = os.path.join(os.getcwd(), "rat")
    download_rat(token, run_version, install_path)
    
def download_rat(token, run_version, install_path):
    '''download the requested version of RAT, unzip it and
    move it all to a rat directory
    Should use method to ship with job instead'''
    cache_path = os.getcwd()
    url = "https://api.github.com/repos/snoplus/rat/tarball/%s" % run_version
    tar_name = 'rat.tar.gz'
    job_tools.download_file(url, token, filename=tar_name, cache_path=cache_path)
    job_tools.untar_file( tarName, installPath, 'rat', cache_path, 1 ) # Strip the first directory


#################################
# Current functions
#################################
def untar_local(tarname):
    '''Untar a repository shipped with the job'''
    install_path = os.getcwd()
    job_tools.untar_file(tarname, installPath, 'rat', os.getcwd(), 0) # nothing to strip, but do need to move
    # For some reason, configure is not x by default
    job_tools.execute('chmod', ['u+x','rat/configure'], cwd=installPath)


def install_local(run_version, env_file):
    '''Install a shipped / downloaded tar for use.
    '''
    #should be in the temporary job directory to do this
    install_path = os.path.join(os.getcwd(), "rat")
    create_env('installerTemp.sh', env_file)
    compile_rat()
    create_env('installerEnv.sh', env_file, install_path)


def create_env(filename, env_file, rat_dir=None):
    '''Create environment file to source for RAT jobs.
    
    If rat_dir is not set, then assumes only base directories are needed.
    '''
    if not os.path.exists(env_file):
        raise Exception("create_env::cannot find environment %s" % env_file)
    fin = file(env_file, 'r')
    env_text = ''
    lines = fin.readlines()
    for i,line in enumerate(lines):
        if 'source ' in line:
            if i==len(lines)-1 or i==len(lines)-2:
                # This is the fixed RAT release line; remove it!
                pass
            else:
                env_text += '%s \n' % line
        else:
            env_text += '%s \n' % line
    if rat_dir is not None:
        env_text += "source %s" % os.path.join(rat_dir, "env.sh")
    env_file = file(filename, 'w')
    env_file.write(env_text)


def compile_rat():
    '''Compile local version of RAT'''
    temp_file = os.path.join(os.getcwd(), 'installerTemp.sh')
    rat_dir   = os.path.join(os.getcwd(), 'rat')
    command_text = "\
#!/bin/bash\n \
source %s\n \
cd %s\n \
./configure \n \
source env.sh\n \
scons" % (temp_file, rat_dir)
    job_tools.execute_complex(command_text)
    #running locally, seems rat is not executable by default!
    job_tools.execute('chmod', ['u+x', 'rat/bin/rat'])


def run_rat(options):
    '''Run RAT according to the options.
    '''
    rat_options = ['-l', 'rat.log']
    if options.input_file:
        rat_options += ['-i', options.input_file]
    if options.output_file:
        rat_options += ['-o', options.output_file]
    if options.n_events:
        rat_options += ['-N', options.n_events]
    elif options.run_time:
        rat_options += ['-T', options.run_time]
    if options.dbuser and options.dbpassword and options.dbname and options.dbprotocol and options.dburl:
        rat_options += ['-b', '%s://%s:%s@%s/%s' % (options.dbprotocol, options.dbuser, options.dbpassword,
                                                    options.dburl, options.dbname)]
    rat_options += [options.rat_macro]
    job_tools.execute_ratenv('rat', rat_options)    
    print os.listdir(os.getcwd())

    
def check_outputs(options):
    '''Before running RAT, check whether output data already exists!
    '''
    # Can only check the outputs for files with output file specified in the command options
    if options.output_file is None:
        print "check_outputs::no output file specified"
        return None
    if options.grid_mode is None:
        print "check_outputs::no grid mode specified"
        return None
    # For output files, check if there is an output root and output ntuple processor in the macro
    ntuple = job_tools.check_processor(options.rat_macro, 'outntuple')
    root = job_tools.check_processor(options.rat_macro, 'outroot')
    suffixes = []
    exists = []
    if root:
        suffixes.append(".root")
        exists.append(False)
    if ntuple:
        suffixes.append(".ntuple.root")
        exists.append(False)
    all_exist = True
    any_exist = False
    if len(suffixes)==0:
        print "check_output::no files to check!"
        print suffixes
        all_exist = False
    for i, s in enumerate(suffixes):
        root_file = "%s%s" % (options.output_file, s)
        lfc_path = os.path.join('lfn:/grid/snoplus.snolab.ca', options.output_dir, root_file)
        exists[i] = False
        try:
            job_tools.exists(lfc_path)
            # lfc exists, get the replica
            exists[i] = True
            any_exist = True
            print "check_outputs::LFN already in use %s" % lfc_path
        except job_tools.JobToolsException as e:
            print "check_outputs::LFN unused %s, %s" % (lfc_path, e)
            all_exist = False
    # Now, if all exist, we need to append the return card and show that the job has in fact completed
    # If only some (i.e. any) exist, but not all, then we need to raise this problem by failing the job!
    dump_out = {}
    if all_exist:
        print "check_outputs::All expected outputs already exist on LFN"
        for s in suffixes:
            try:
                root_file = "%s%s" % (options.output_file, s)
                lfc_path = os.path.join('lfn:/grid/snoplus.snolab.ca', options.output_dir, root_file)
                guid = job_tools.getguid(lfc_path)
                replica = job_tools.listreps(lfc_path)[0]
                checksum = job_tools.checksum(replica)
                size = job_tools.getsize(replica)
                
                # Note that the actual se and replica information may be different to the local information
                se_name = replica
                if se_name.startswith('srm://'):
                    se_name = se_name[6:]
                se_name = se_name.split('/')[0] # never any /
                dump_out[root_file] = {}
                dump_out[root_file]['guid'] = guid
                dump_out[root_file]['se'] = replica
                dump_out[root_file]['size'] = size
                dump_out[root_file]['cksum'] = checksum
                dump_out[root_file]['name'] = se_name
                dump_out[root_file]['lfc'] = lfc_path
            except job_tools.JobToolsException as e:
                print "check_outputs::Problem checking output for %s: %s" % (lfc_path, e)
                raise
        return dump_out
    else:
        return None


def check_output_file(options):  
    '''Check root file for TTree entries
    '''
    # Can only check the outputs for files with output file specified in the command options
    if options.output_file is None:
        print "check_output_file::no output file specified"
        return None

    # Can only check the outputs for files with number of events specified in the command options
    if options.n_events is None:
        print "check_output_file::no number of events specified"
        return None

    # Create command to run python script to check TTree entries
    # Will raise an exception if the checks fail
    ntuple = job_tools.check_processor(options.rat_macro, 'outntuple')
    root = job_tools.check_processor(options.rat_macro, 'outroot')
    soc = job_tools.check_processor(options.rat_macro, 'outsoc')
    if root or soc:
        rtc, out, err = job_tools.execute_ratenv('python', ['check_root_output.py', '-f', '%s.root' % options.output_file,
                                                        '-n', options.n_events, '-v', options.base_version])
        if rtc==0:
            # this must be true (exception raised otherwise)
            print "Outputs file checks successful"
    if ntuple:
        rtc, out, err = job_tools.execute_ratenv('python', ['check_root_output.py', '-f', '%s.ntuple.root' % options.output_file,
                                                        '-n', options.n_events, '-v', options.base_version])
        if rtc==0:
            # this must be true (exception raised otherwise)
            print "Outputs file checks successful"


def copy_data(output_dir, grid_mode):
    '''Copy all output data (i.e. any .root output files)
    grid_mode true: use lcg-cr to copy and register file
             false: use cp (the job runs in a temp dir, any files left at end are deleted eventually)
    '''
    # Still need catch-all files here (in case ROOT splits them)
    root_files = glob.glob('*.root')
    dump_out = job_tools.copy_data(root_files, output_dir, grid_mode)
    return dump_out


if __name__ == '__main__':
    print 'ratRunner...??'
    parser = optparse.OptionParser( usage = "ganga %prog [flags]")
    parser.add_option("-t", dest="token", help="Token to use, if in token mode")
    parser.add_option("-b", dest="base_version", help="RAT base version")
    parser.add_option("-v", dest="run_version", help="RAT version to run")
    parser.add_option("-e", dest="env_file", help="Software environment file to source")
    parser.add_option("-m", dest="rat_macro", help="RAT macro to use")
    parser.add_option("-d", dest="output_dir", help="LFC/SURL relative directory if in grid mode, full dir path if not in grid mode")
    parser.add_option("-g", dest="grid_mode", default=None, help="Grid mode (allowed modes are srm and lcg)")
    parser.add_option("-f", dest="zip_filename", help="Zip filename for output file")
    parser.add_option("-i", dest="input_file", default=None, help="Specify input file")
    parser.add_option("-o", dest="output_file", default=None, help="Specify output file")
    parser.add_option("-N", dest="n_events", default=None, help="Number of events (must not be in macro)")
    parser.add_option("-T", dest="run_time", default=None, help="Duration of run (cannot use with n_events)")
    parser.add_option("--dbuser", dest="dbuser", default=None, help="Database user")
    parser.add_option("--dbpassword", dest="dbpassword", default=None, help="Database password")
    parser.add_option("--dbname", dest="dbname", default=None, help="Database name")
    parser.add_option("--dbprotocol", dest="dbprotocol", default=None, help="Database protocol (http or https)")
    parser.add_option("--dburl", dest="dburl", default=None, help="Database URL (sans protocol)")
    parser.add_option("--nostore", dest="nostore", action="store_true", help="Don't copy the outputs at the end")
    parser.add_option("--voproxy", dest="voproxy", default=None, help="VO proxy location, MUST be used with grid srm mode")

    (options, args) = parser.parse_args()
    if not options.base_version:
        print ' need base version'
    if not options.run_version:
        print 'RUNNING WITH A FIXED RELEASE'
    if not options.rat_macro:
        print ' need macro'
    if not options.output_dir:
        print ' need output directory'
    if not options.env_file:
        print ' need environment file'
    if options.grid_mode=='srm':
        if not options.voproxy:
            print 'Grid %s mode, must define voproxy' % options.grid_mode
            parser.print_help()
            raise Exception
        elif not os.path.exists(options.voproxy):
            print 'Grid %s mode, must define valid voproxy' % options.grid_mode
            parser.print_help()
            raise Exception
    if not options.base_version or not options.rat_macro or not options.output_dir or not options.env_file:
        print 'options not all present, cannot run'
        parser.print_help()
        raise Exception
    elif options.run_version and not options.token and not options.zip_filename:
        print 'must choose one of token/filename to run with a specific hash version'
        parser.print_help()
        raise Exception
    elif options.token and options.zip_filename:
        print 'must choose only one of token/filename'
        parser.print_help()
        raise Exception

    # Now we're sure that we can run

    # First, download any software and setup the environment
    # that will be used by all subsequent functions
    job_helper = job_tools.JobHelper.get_instance() # Singleton
    if options.run_version:
        if options.token:
            download_local(options.token, options.run_version)
        else:
            untar_local(options.zip_filename)
        # all args have to be strings anyway (from Ganga) - force base version into a string
        install_local(options.run_version, options.env_file)
        job_helper.add_environment(os.path.join(os.getcwd(), "installerEnv.sh"))
    else:
        # Still need to ensure that the JobHelper singleton knows about the environment
        job_helper.add_environment(options.env_file)

    # Set the default environments that are needed
    if options.grid_mode=='srm':
        job_helper.set_variables(X509_USER_PROXY=options.voproxy)

    dbAccess = None
    if options.dbuser and options.dbpassword and options.dbname and options.dbprotocol and options.dburl:
        dbAccess = {}
        dbAccess['user'] = options.dbuser
        dbAccess['password'] = options.dbpassword
        dbAccess['name'] = options.dbname
        dbAccess['protocol'] = options.dbprotocol
        dbAccess['url'] = options.dburl
    
    # Check if the output already exists
    dump_out = check_outputs(options)
    if dump_out is None:
        # We do need to run the job
        run_rat(options)
        check_output_file(options)
        dump_out = {}
        if not options.nostore:
             # the nostore option is only applied when running in testing mode for production
            dump_out = copy_data(options.output_dir, options.grid_mode)
        else:
            # return a dummy output dump - no storage information but the production scripts expect it
            dump_out = {"DATA": "DELETED"}
        return_card = file('return_card.js','w')
        return_card.write(json.dumps(dump_out))
        return_card.close()
    else:
        return_card = file('return_card.js','w')
        return_card.write(json.dumps(dump_out))
        return_card.close()
        raise Exception("RATRUNNER: requested outputs already exist, see return_card.js for details")
            



from __future__ import absolute_import

from Ganga.GPIDev.Base.Proxy import stripProxy, implRef
from Ganga.Utility.logging import getLogger

class RegistrySliceProxy(object):

    """This object is an access list to registry slices"""

    def __init__(self, impl):
        setattr(self, implRef, impl)

    def ids(self, minid=None, maxid=None):
        """ Return a list of ids of all objects.
        """
        return stripProxy(self).ids(minid, maxid)

    def clean(self, confirm=False, force=False):
        """ Cleans this registry completely.
        Returns True on success, False on failure"""
        return stripProxy(self).clean(confirm, force)

    def incomplete_ids(self):
        try:
            return stripProxy(self).objects._incomplete_objects
        except Exception as x:
            return []

    def __iter__(self):
        """ Looping. Example:
        for j in jobs:
          print(j.id)
        """
        class Iterator(object):

            def __init__(self, reg):
                self.it = stripProxy(reg).__iter__()

            def __iter__(self): return self

            def next(self):
                return _wrap(next(self.it))
        return Iterator(self)

    def __contains__(self, j):
        return stripProxy(self).__contains__(stripProxy(j))

    def __len__(self):
        return stripProxy(self).__len__()

    def __call__(self, arg):
        return _wrap(stripProxy(self).__call__(arg))

    def select(self, minid=None, maxid=None, **attrs):
        """ Select a subset of objects. Examples for jobs:
        jobs.select(10): select jobs with ids higher or equal to 10;
        jobs.select(10,20) select jobs with ids in 10,20 range (inclusive);
        jobs.select(status='new') select all jobs with new status;
        jobs.select(name='some') select all jobs with some name;
        jobs.select(application='Executable') select all jobs with Executable application;
        jobs.select(backend='Local') select all jobs with Local backend.
        """
        unwrap_attrs = {}
        for a in attrs:
            unwrap_attrs[a] = _unwrap(attrs[a])
        logger = getLogger()
        logger.debug("Calling: %s" % str(stripProxy(self).select))
        return self.__class__(stripProxy(self).select(minid, maxid, **unwrap_attrs))

    def _display(self, interactive=True):
        return stripProxy(self)._display(interactive)

    def __str__(self):
        return self._display(interactive=0)

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('registry...')
            return
        p.text(self._display())


# wrap Proxy around a ganga object (or a list of ganga objects)
# leave all others unchanged
from Ganga.GPIDev.Base.Proxy import addProxy
from Ganga.GPIDev.Base.Objects import GangaObject
from .RegistrySlice import RegistrySlice

from Ganga.GPIDev.Base.Proxy import isType

def _wrap(obj):
    if isType(obj, GangaObject):
        return addProxy(obj)
    if isType(obj, RegistrySlice):
        return obj._proxyClass(obj)
    if isType(obj, list):
        return map(addProxy, obj)
    return obj

# strip Proxy and get into the ganga object implementation

def _unwrap(obj):
    return stripProxy(obj)




"""Module containing class for handling job submission to Remote backend"""

__author__ = "Mark Slater <mws@hep.ph.bham.ac.uk>"
__date__ = "10 June 2008"
__version__ = "1.0"

from Ganga.Core import Sandbox
from Ganga.GPIDev.Adapters.IBackend import IBackend
from Ganga.GPIDev.Lib.File.FileBuffer import FileBuffer
from Ganga.GPIDev.Schema import ComponentItem, Schema, SimpleItem, Version

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()

import inspect
import os
from Ganga.Lib.Root import randomString


def shutdown_transport(tr):
    """Shutdown the transport cleanly - otherwise python 2.5 throws a wobble"""
    if (tr != None):
        tr.close()
        tr = None


class Remote(IBackend):

    """Remote backend - submit jobs to a Remote pool.

    The remote backend works as an SSH tunnel to a remote site
    where a ganga session is opened and the job submitted there
    using the specified remote_backend. It is (in theory!)
    transparent to the user and should allow submission of any jobs
    to any backends that are already possible in Ganga.

    NOTE: Due to the file transfers required, there can be some slow
    down during submission and monitoring


    E.g. 1 - Hello World example submitted to local backend:

    j = Job(application=Executable(exe='/bin/echo',args=['Hello World']), backend="Remote")
    j.backend.host = "bluebear.bham.ac.uk"                  # Host name
    j.backend.username = "slatermw"                         # User name
    j.backend.ganga_cmd = "/bb/projects/Ganga/runGanga"     # Ganga Command line on remote site
    j.backend.ganga_dir = "/bb/phy/slatermw/gangadir/remote_jobs"  # Where to store the jobs
    j.backend.remote_backend = Local()
    j.submit()


    E.g. 2 - Root example submitted to PBS backend:

    r = Root()
    r.version = '5.14.00'
    r.script = 'gengaus.C'

    j = Job(application=r,backend="Remote")
    j.backend.host = "bluebear.bham.ac.uk"
    j.backend.username = "slatermw"
    j.backend.ganga_cmd = "/bb/projects/Ganga/runGanga"
    j.backend.ganga_dir = "/bb/phy/slatermw/gangadir/remote_jobs"
    j.outputsandbox = ['gaus.txt']
    j.backend.remote_backend = PBS()
    j.submit()


    E.g. 3 - Athena example submitted to LCG backend
    NOTE: you don't need a grid certificate (or UI) available on the local machine,
    just the remote machine

    j = Job()
    j.name='Ex3_2_1'
    j.application=Athena()
    j.application.prepare(athena_compile=False)
    j.application.option_file='/disk/f8b/home/mws/athena/testarea/13.0.40/PhysicsAnalysis/AnalysisCommon/UserAnalysis/run/AthExHelloWorld_jobOptions.py'

    j.backend = Remote()
    j.backend.host = "bluebear.bham.ac.uk"
    j.backend.username = "slatermw"
    j.backend.ganga_cmd = "/bb/projects/Ganga/runGanga"
    j.backend.ganga_dir = "/bb/phy/slatermw/gangadir/remote_jobs"   
    j.backend.environment = {'ATLAS_VERSION' : '13.0.40'}     # Additional environment variables
    j.backend.remote_backend = LCG()
    j.backend.remote_backend.CE = 'epgce2.ph.bham.ac.uk:2119/jobmanager-lcgpbs-short'

    j.submit()

    E.g. 4 - Hello World submitted at CERN on LSF using atlas startup

    j = Job()
    j.backend = Remote()
    j.backend.host = "lxplus.cern.ch"
    j.backend.username = "mslater"
    j.backend.ganga_cmd = "ganga"
    j.backend.ganga_dir = "/afs/cern.ch/user/m/mslater/gangadir/remote_jobs"
    j.backend.pre_script = ['source /afs/cern.ch/sw/ganga/install/etc/setup-atlas.csh'] # source the atlas setup script before running ganga
    j.backend.remote_backend = LSF()
    j.submit()

    """

    _schema = Schema(Version(1, 0), {
        "remote_backend": ComponentItem('backends', doc='specification of the resources to be used (e.g. batch system)'),
        "host": SimpleItem(defvalue="", doc="The remote host and port number ('host:port') to use. Default port is 22."),
        "ssh_key": SimpleItem(defvalue="", doc="Set to true to the location of the the ssh key to use for authentication, e.g. /home/mws/.ssh/id_rsa. Note, you should make sure 'key_type' is also set correctly."),
        "key_type": SimpleItem(defvalue="RSA", doc="Set to the type of ssh key to use (if required). Possible values are 'RSA' and 'DSS'."),
        "username": SimpleItem(defvalue="", doc="The username at the remote host"),
        "ganga_dir": SimpleItem(defvalue="", doc="The directory to use for the remote workspace, repository, etc."),
        "ganga_cmd": SimpleItem(defvalue="", doc="Command line to start ganga on the remote host"),
        "environment": SimpleItem(defvalue={}, doc="Overides any environment variables set in the job"),
        "pre_script": SimpleItem(defvalue=[''], doc="Sequence of commands to execute before running Ganga on the remote site"),
        'remote_job_id': SimpleItem(defvalue=0, protected=1, copyable=0, doc='Remote job id.'),
        'exitcode': SimpleItem(defvalue=0, protected=1, copyable=0, doc='Application exit code'),
        'actualCE': SimpleItem(defvalue=0, protected=1, copyable=0, doc='Computing Element where the job actually runs.')
    })

    _category = "backends"
    _name = "Remote"
    #_hidden = False # KUBA: temporarily disabled from the public
    _port = 22
    _transport = None
    _sftp = None
    _code = randomString()
    _transportarray = None
    _key = {}

    _exportmethods = ['setup']

    def __init__(self):
        super(Remote, self).__init__()

    def __del__(self):
        if (self._transport != None):
            self._transport.close()
            self._transport = None

    def setup(self):  # KUBA: generic setup hook
        job = self.getJobObject()
        if job.status in ['submitted', 'running', 'completing']:

            # Send a script over to the remote site that updates this jobs
            # info with the info of the remote job
            import os

            # Create a ganga script that updates the job info from the remote
            # site
            script = """#!/usr/bin/env python
from __future__ import print_function
#-----------------------------------------------------
# This is a setup script for a remote job. It
# does very litte
#-----------------------------------------------------

# print a finished token
print("***_FINISHED_***")
"""

            # check for the connection
            if (self.opentransport() == False):
                return False

            # send the script
            #script_name = '/__setupscript__%s.py' % self._code
            #self._sftp.open(self.ganga_dir + script_name, 'w').write(script)

            # run the script
            #stdout, stderr = self.run_remote_script( script_name, self.pre_script )

            # remove the script
            #self._sftp.remove(self.ganga_dir + script_name)

        return True

    def opentransport(self):

        import paramiko
        import getpass
        import atexit

        if (self._transport != None):
            # transport is open
            return

        # check for a useable transport for this username and host
        if Remote._transportarray != None:
            for t in Remote._transportarray:
                if (t != None) and (t[0] == self.username) and (t[1] == self.host):

                    # check for too many retries on the same host
                    if t[2] == None or t[3] == None:
                        logger.warning("Too many retries for remote host " + self.username +
                                       "@" + self.host + ". Restart Ganga to have another go.")
                        return False

                    self._transport = t[2]
                    self._sftp = t[3]

                    # ensure that the remote dir is still there - it will crash if the dir structure
                    # changes with the sftp sill open
                    channel = self._transport.open_session()
                    channel.exec_command('mkdir -p ' + self.ganga_dir)
                    bufout = ""
                    while not channel.exit_status_ready():
                        if channel.recv_ready():
                            bufout = channel.recv(1024)

                    return

        # Ask user for password - give three tries
        num_try = 0
        password = ""
        while num_try < 3:

            try:
                temp_host = self.host
                temp_port = self._port
                if self.host.find(":") != -1:
                    # user specified port
                    temp_port = eval(self.host[self.host.find(":") + 1:])
                    temp_host = self.host[: self.host.find(":")]

                self._transport = paramiko.Transport((temp_host, temp_port))

                # avoid hang on exit my daemonising the thread
                self._transport.setDaemon(True)

                # register for proper shutdown
                atexit.register(shutdown_transport, self._transport)

                if self.ssh_key != "" and os.path.exists(os.path.expanduser(os.path.expandvars(self.ssh_key))):
                    privatekeyfile = os.path.expanduser(
                        os.path.expandvars(self.ssh_key))

                    if self.ssh_key not in Remote._key:

                        if self.key_type == "RSA":
                            password = getpass.getpass(
                                'Enter passphrase for key \'%s\': ' % (self.ssh_key))
                            Remote._key[self.ssh_key] = paramiko.RSAKey.from_private_key_file(
                                privatekeyfile, password=password)
                        elif self.key_type == "DSS":
                            password = getpass.getpass(
                                'Enter passphrase for key \'%s\': ' % (self.ssh_key))
                            Remote._key[self.ssh_key] = paramiko.DSSKey.from_private_key_file(
                                privatekeyfile, password=password)
                        else:
                            logger.error(
                                "Unknown ssh key_type '%s'. Unable to connect." % self.key_type)
                            return False

                    self._transport.connect(
                        username=self.username, pkey=Remote._key[self.ssh_key])
                else:
                    logger.debug("SSH key: %s" % self.ssh_key)
                    if os.path.exists(os.path.expanduser(os.path.expandvars(self.ssh_key))):
                        logger.debug(
                            "PATH: %s Exists" % os.path.expanduser(os.path.expandvars(self.ssh_key)))
                    else:
                        logger.debug("PATH: %s Does NOT Exist" % os.path.expanduser(
                            os.path.expandvars(self.ssh_key)))

                    if self.username != "" and self.host != "":
                        password = getpass.getpass(
                            'Password for %s@%s: ' % (self.username, self.host))
                        self._transport.connect(
                            username=self.username, password=password)
                    elif self.username == "":
                        logger.error("ERROR: USERNAME NOT DEFINED!!!")
                        return False
                    elif self.host == "":
                        logger.error("ERROR: HOSTNAME NOT DEFINED!!!")
                        return False
                    else:
                        pass

                # blank the password just in case
                password = "                                                "

                channel = self._transport.open_session()
                channel.exec_command('mkdir -p ' + self.ganga_dir)
                self._sftp = paramiko.SFTPClient.from_transport(
                    self._transport)

                # Add to the transport array
                Remote._transportarray = [Remote._transportarray,
                                          [self.username, self.host, self._transport, self._sftp]]
                num_try = 1000

            except Exception as err:
                logger.debug("Err: %s" %str(err))
                logger.warning("Error when comunicating with remote host. Retrying...")
                self._transport = None
                self._sftp = None
                if self.ssh_key in Remote._key:
                    del Remote._key[self.ssh_key]

            num_try = num_try + 1

        if num_try == 3:
            logger.error("Could not logon to remote host " + self.username + "@" +
                         self.host + " after three attempts. Restart Ganga to have another go.")
            Remote._transportarray = [Remote._transportarray,
                                      [self.username, self.host, None, None]]
            return False

        return True

    def run_remote_script(self, script_name, pre_script):
        """Run a ganga script on the remote site"""

        import getpass

        # Set up a command file to source. This gets around a silly alias
        # problem
        cmd_str = ""
        for c in pre_script:
            cmd_str += c + '\n'

        cmd_str += self.ganga_cmd + \
            " -o\'[Configuration]gangadir=" + self.ganga_dir + "\' "
        cmd_str += self.ganga_dir + script_name + '\n'
        cmd_file = os.path.join(
            self.ganga_dir, "__gangacmd__" + randomString())
        self._sftp.open(cmd_file, 'w').write(cmd_str)

        # run ganga command
        channel = self._transport.open_session()
        channel.exec_command("source " + cmd_file)

        # Read the output after command
        stdout = bufout = ""
        stderr = buferr = ""
        grid_ok = False

        while not channel.exit_status_ready():

            if channel.recv_ready():
                bufout = channel.recv(1024)
                stdout += bufout

            if channel.recv_stderr_ready():
                buferr = channel.recv_stderr(1024)
                stderr += buferr

            if stdout.find("***_FINISHED_***") != -1:
                break

            if (bufout.find("GRID pass") != -1 or buferr.find("GRID pass") != -1):
                grid_ok = True
                password = getpass.getpass('Enter GRID pass phrase: ')
                channel.send(password + "\n")
                password = ""

            bufout = buferr = ""

        self._sftp.remove(cmd_file)

        return stdout, stderr

    def submit(self, jobconfig, master_input_sandbox):
        """Submit the job to the remote backend.

            Return value: True if job is submitted successfully,
                          or False otherwise"""

        import os
        import getpass

        # First some sanity checks...
        fail = 0
        if self.remote_backend == None:
            logger.error("No backend specified for remote host.")
            fail = 1
        if self.host == "":
            logger.error("No remote host specified.")
            fail = 1
        if self.username == "":
            logger.error("No username specified.")
            fail = 1
        if self.ganga_dir == "":
            logger.error("No remote ganga directory specified.")
            fail = 1
        if self.ganga_cmd == "":
            logger.error("No ganga command specified.")
            fail = 1

        if fail:
            return 0

        # initiate the connection
        if self.opentransport() == False:
            return 0

        # Tar up the input sandbox and copy to the remote cluster
        job = self.getJobObject()
        subjob_input_sandbox = job.createPackedInputSandbox(
            jobconfig.getSandboxFiles())
        input_sandbox = subjob_input_sandbox + master_input_sandbox

        # send the sandbox
        sbx_name = '/__subjob_input_sbx__%s' % self._code
        self._sftp.put(subjob_input_sandbox[0], self.ganga_dir + sbx_name)
        sbx_name = '/__master_input_sbx__%s' % self._code
        self._sftp.put(master_input_sandbox[0], self.ganga_dir + sbx_name)

        # run the submit script on the remote cluster
        scriptpath = self.preparejob(jobconfig, master_input_sandbox)

        # send the script
        data = open(scriptpath, 'r').read()
        script_name = '/__jobscript_run__%s.py' % self._code
        self._sftp.open(self.ganga_dir + script_name, 'w').write(data)

        # run the script
        stdout, stderr = self.run_remote_script(script_name, self.pre_script)

        # delete the jobscript
        self._sftp.remove(self.ganga_dir + script_name)

        # Copy the job object
        if stdout.find("***_FINISHED_***") != -1:
            status, outputdir, id, be = self.grabremoteinfo(stdout)

            self.remote_job_id = id
            if hasattr(self.remote_backend, 'exitcode'):
                self.exitcode = be.exitcode
            if hasattr(self.remote_backend, 'actualCE'):
                self.actualCE = be.actualCE

            # copy each variable in the schema
            # Please can someone tell me why I can't just do
            # self.remote_backend = be?
            for o in be._schema.allItems():
                exec("self.remote_backend." + o[0] + " = be." + o[0])

            return 1
        else:
            logger.error("Problem submitting the job on the remote site.")
            logger.error("<last 1536 bytes of stderr>")
            cut = stderr[len(stderr) - 1536:]

            for ln in cut.splitlines():
                logger.error(ln)

            logger.error("<end of last 1536 bytes of stderr>")

        return 0

    def kill(self):
        """Kill running job.

           No arguments other than self

           Return value: True if job killed successfully,
                         or False otherwise"""

        script = """#!/usr/bin/env python
from __future__ import print_function
#-----------------------------------------------------
# This is a kill script for a remote job. It
# attempts to kill the given job and returns
#-----------------------------------------------------
import os,os.path,shutil,tempfile
import sys,popen2,time,traceback

############################################################################################

###INLINEMODULES###

############################################################################################

code = ###CODE###
jid = ###JOBID###

j = jobs( jid )
j.kill()

# Start pickle token
print("***_START_PICKLE_***")

# pickle the job
import pickle
print(j.outputdir)
print(pickle.dumps(j._impl))
print(j)

# print a finished token
print("***_END_PICKLE_***")
print("***_FINISHED_***")
"""

        script = script.replace('###CODE###', repr(self._code))
        script = script.replace('###JOBID###', str(self.remote_job_id))

        # check for the connection
        if (self.opentransport() == False):
            return 0

        # send the script
        script_name = '/__jobscript_kill__%s.py' % self._code
        self._sftp.open(self.ganga_dir + script_name, 'w').write(script)

        # run the script
        stdout, stderr = self.run_remote_script(script_name, self.pre_script)

        # Copy the job object
        if stdout.find("***_FINISHED_***") != -1:
            status, outputdir, id, be = self.grabremoteinfo(stdout)

            if status == 'killed':
                return True

        return False

    def remove(self):
        """Remove the selected job from the remote site

           No arguments other than self

           Return value: True if job removed successfully,
                         or False otherwise"""

        script = """#!/usr/bin/env python
from __future__ import print_function
#-----------------------------------------------------
# This is a remove script for a remote job. It
# attempts to kill the given job and returns
#-----------------------------------------------------
import os,os.path,shutil,tempfile
import sys,popen2,time,traceback

############################################################################################

###INLINEMODULES###

############################################################################################

code = ###CODE###
jid = ###JOBID###

j = jobs( jid )
j.remove()

jobs( jid )

# print a finished token
print("***_FINISHED_***")
"""

        script = script.replace('###CODE###', repr(self._code))
        script = script.replace('###JOBID###', str(self.remote_job_id))

        # check for the connection
        if (self.opentransport() == False):
            return 0

        # send the script
        script_name = '/__jobscript_remove__%s.py' % self._code
        self._sftp.open(self.ganga_dir + script_name, 'w').write(script)

        # run the script
        stdout, stderr = self.run_remote_script(script_name, self.pre_script)

        # Copy the job object
        if stdout.find("***_FINISHED_***") != -1:
            return True

        return False

    def resubmit(self):
        """Resubmit the job.

           No arguments other than self

           Return value: 1 if job was resubmitted,
                         or 0 otherwise"""

        script = """#!/usr/bin/env python
from __future__ import print_function
#-----------------------------------------------------
# This is a resubmit script for a remote job. It
# attempts to kill the given job and returns
#-----------------------------------------------------
import os,os.path,shutil,tempfile
import sys,popen2,time,traceback

############################################################################################

###INLINEMODULES###

############################################################################################

code = ###CODE###
jid = ###JOBID###

j = jobs( jid )
j.resubmit()

# Start pickle token
print("***_START_PICKLE_***")

# pickle the job
import pickle
print(j.outputdir)
print(pickle.dumps(j._impl))
print(j)

# print a finished token
print("***_END_PICKLE_***")
print("***_FINISHED_***")
"""

        script = script.replace('###CODE###', repr(self._code))
        script = script.replace('###JOBID###', str(self.remote_job_id))

        # check for the connection
        if (self.opentransport() == False):
            return 0

        # send the script
        script_name = '/__jobscript_resubmit__%s.py' % self._code
        self._sftp.open(self.ganga_dir + script_name, 'w').write(script)

        # run the script
        stdout, stderr = self.run_remote_script(script_name, self.pre_script)

        # Copy the job object
        if stdout.find("***_FINISHED_***") != -1:
            status, outputdir, id, be = self.grabremoteinfo(stdout)

            if status == 'submitted' or status == 'running':
                return 1

        return 0

    def grabremoteinfo(self, out):

        import pickle

        # Find the start and end of the pickle
        start = out.find("***_START_PICKLE_***") + len("***_START_PICKLE_***")
        stop = out.find("***_END_PICKLE_***")
        outputdir = out[start + 1:out.find("\n", start + 1) - 1]
        pickle_str = out[out.find("\n", start + 1) + 1:stop]

        # Now unpickle and recreate the job
        j = pickle.loads(pickle_str)

        return j.status, outputdir, j.id, j.backend

    def preparejob(self, jobconfig, master_input_sandbox):
        """Prepare the script to create the job on the remote host"""

        import tempfile

        workdir = tempfile.mkdtemp()
        job = self.getJobObject()

        script = """#!/usr/bin/env python
from __future__ import print_function
#-----------------------------------------------------
# This job wrapper script is automatically created by
# GANGA Remote backend handler.
#
# It controls:
# 1. unpack input sandbox
# 2. create the new job
# 3. submit it
#-----------------------------------------------------
import os,os.path,shutil,tempfile
import sys,popen2,time,traceback
import tarfile

############################################################################################

###INLINEMODULES###

############################################################################################

j = Job()

output_sandbox = ###OUTPUTSANDBOX###
input_sandbox = ###INPUTSANDBOX###
appexec = ###APPLICATIONEXEC###
appargs = ###APPLICATIONARGS###
back_end = ###BACKEND###
ganga_dir = ###GANGADIR###
code = ###CODE###
environment = ###ENVIRONMENT###
user_env = ###USERENV###

if user_env != None:
   for env_var in user_env:
      environment[env_var] = user_env[env_var]

j.outputsandbox = output_sandbox
j.backend = back_end

# Unpack the input sandboxes
shutil.move(os.path.expanduser(ganga_dir + "/__subjob_input_sbx__" + code), j.inputdir+"/__subjob_input_sbx__")
shutil.move(os.path.expanduser(ganga_dir + "/__master_input_sbx__" + code), j.inputdir+"/__master_input_sbx__")

# Add the files in the sandbox to the job
inputsbx = []
fullsbxlist = []
try:
   tar = tarfile.open(j.inputdir+"/__master_input_sbx__")
   filelist = tar.getnames()
   print(filelist)
   
   for f in filelist:
      fullsbxlist.append( f )
      inputsbx.append( j.inputdir + "/" + f )

except:
   print("Unable to open master input sandbox")

try:
   tar = tarfile.open(j.inputdir+"/__subjob_input_sbx__")
   filelist = tar.getnames()

   for f in filelist:
      fullsbxlist.append( f )
      inputsbx.append( j.inputdir + "/" + f )

except:
   print("Unable to open subjob input sandbox")

# sort out the path of the exe
if appexec in fullsbxlist:
   j.application = Executable ( exe = File(os.path.join(j.inputdir, appexec)), args = appargs, env = environment )
   print("Script found: %s" % appexec)
else:
   j.application = Executable ( exe = appexec, args = appargs, env = environment )

   
j.inputsandbox = inputsbx

getPackedInputSandbox(j.inputdir+"/__subjob_input_sbx__", j.inputdir + "/.")
getPackedInputSandbox(j.inputdir+"/__master_input_sbx__", j.inputdir + "/.")

# submit the job
j.submit()

# Start pickle token
print("***_START_PICKLE_***")

# pickle the job
import pickle
print(j.outputdir)
print(pickle.dumps(j._impl))

# print a finished token
print("***_END_PICKLE_***")
print("***_FINISHED_***")
"""
        import inspect
        import Ganga.Core.Sandbox as Sandbox
        script = script.replace('###ENVIRONMENT###', repr(jobconfig.env))
        script = script.replace('###USERENV###', repr(self.environment))
        script = script.replace(
            '###INLINEMODULES###', inspect.getsource(Sandbox.WNSandbox))
        script = script.replace(
            '###OUTPUTSANDBOX###', repr(jobconfig.outputbox))
        script = script.replace(
            '###APPLICATIONEXEC###', repr(os.path.basename(jobconfig.getExeString())))
        script = script.replace(
            '###APPLICATIONARGS###', repr(jobconfig.getArgStrings()))

        # get a string describing the required backend
        import cStringIO
        be_out = cStringIO.StringIO()
        job.backend.remote_backend.printTree(be_out, "copyable")
        be_str = be_out.getvalue()
        script = script.replace('###BACKEND###', be_str)

        script = script.replace('###GANGADIR###', repr(self.ganga_dir))
        script = script.replace('###CODE###', repr(self._code))

        sandbox_list = jobconfig.getSandboxFiles()

        str_list = "[ "
        for fname in sandbox_list:
            str_list += "j.inputdir + '/' + " + \
                repr(os.path.basename(fname.name))
            str_list += ", "

        str_list += "j.inputdir + '/__master_input_sbx__' ]"

        script = script.replace('###INPUTSANDBOX###', str_list)
        return job.getInputWorkspace().writefile(FileBuffer('__jobscript__.py', script), executable=0)

    @staticmethod
    def updateMonitoringInformation(jobs):

        # Send a script over to the remote site that updates this jobs
        # info with the info of the remote job
        import os
        import getpass

        # first, loop over the jobs and sort by host, username, gangadir and
        # pre_script
        jobs_sort = {}
        for j in jobs:
            host_str = j.backend.username + "@" + j.backend.host + ":" + \
                j.backend.ganga_dir + "+" + ';'.join(j.backend.pre_script)
            if host_str not in jobs_sort:
                jobs_sort[host_str] = []

            jobs_sort[host_str].append(j)

        for host_str in jobs_sort:
            # Create a ganga script that updates the job info for all jobs at
            # this remote site
            script = """#!/usr/bin/env python
from __future__ import print_function
#-----------------------------------------------------
# This is a monitoring script for a remote job. It
# outputs some useful job info and exits
#-----------------------------------------------------
import os,os.path,shutil,tempfile
import sys,popen2,time,traceback

############################################################################################

###INLINEMODULES###

############################################################################################

code = ###CODE###
jids = ###JOBID###

runMonitoring()

import pickle

for jid in jids:

    j = jobs( jid )

    # Start pickle token
    print("***_START_PICKLE_***")

    # pickle the job
    print(j.outputdir)
    print(pickle.dumps(j._impl))
    print(j)

    # print a finished token
    print("***_END_PICKLE_***")

print("***_FINISHED_***")
"""

            mj = jobs_sort[host_str][0]
            script = script.replace('###CODE###', repr(mj.backend._code))
            rem_ids = []
            for j in jobs_sort[host_str]:
                rem_ids.append(j.backend.remote_job_id)
            script = script.replace('###JOBID###', str(rem_ids))

            # check for the connection
            if (mj.backend.opentransport() == False):
                return 0

            # send the script
            script_name = '/__jobscript__%s.py' % mj.backend._code
            mj.backend._sftp.open(
                mj.backend.ganga_dir + script_name, 'w').write(script)

            # run the script
            stdout, stderr = mj.backend.run_remote_script(
                script_name, mj.backend.pre_script)

            # Copy the job object
            if stdout.find("***_FINISHED_***") != -1:

                start_pos = stdout.find("***_START_PICKLE_***")
                end_pos = stdout.find(
                    "***_END_PICKLE_***") + len("***_END_PICKLE_***")

                while start_pos != -1 and end_pos != -1:
                    pickle_str = stdout[start_pos:end_pos + 1]

                    status, outputdir, id, be = mj.backend.grabremoteinfo(
                        pickle_str)

                    # find the job and update it
                    found = False
                    for j in jobs_sort[host_str]:

                        if (id == j.backend.remote_job_id):
                            found = True
                            if status != j.status:
                                j.updateStatus(status)

                            if hasattr(j.backend.remote_backend, 'exitcode'):
                                j.backend.exitcode = be.exitcode
                            if hasattr(j.backend.remote_backend, 'actualCE'):
                                j.backend.actualCE = be.actualCE

                            for o in be._schema.allItems():
                                exec(
                                    "j.backend.remote_backend." + o[0] + " = be." + o[0])

                            # check for completed or failed and pull the output
                            # if required
                            if j.status == 'completed' or j.status == 'failed':

                                # we should have output, so get the file list
                                # first
                                filelist = j.backend._sftp.listdir(outputdir)

                                # go through and sftp them back
                                for fname in filelist:
                                    data = j.backend._sftp.open(
                                        outputdir + '/' + fname, 'r').read()
                                    open(
                                        j.outputdir + '/' + os.path.basename(fname), 'w').write(data)

                    if not found:
                        logger.warning(
                            "Couldn't match remote id %d with monitored job. Serious problems in Remote monitoring." % id)

                    start_pos = stdout.find("***_START_PICKLE_***", end_pos)
                    end_pos = stdout.find(
                        "***_END_PICKLE_***", end_pos) + len("***_END_PICKLE_***")

            # remove the script
            j.backend._sftp.remove(j.backend.ganga_dir + script_name)

        return None




from GangaTest.Framework.tests import GangaGPITestCase
#from unittest import TestCase
from xml.dom import minidom

from GangaRobot.Lib.Base.Extract import Node

def equal(node1, node2):
    """Return True if node1 equals node2, otherwise False.
    
    Handles type Node by comparing result of toxml().
    Handles list types by comparing elements.
    Handles other types by == comparison.
    
    """
    if isinstance(node1, list) and isinstance(node2, list):
        if len(node1) != len(node2):
            return False
        else:
            for n1, n2 in zip(node1, node2):
                if not equal(n1, n2):
                    return False
    if isinstance(node1, Node) and isinstance(node2, Node):
        return node1.toxml() == node2.toxml()
    else:
        return node1 == node2


class TestExtract(GangaGPITestCase):

    """Tests of extracted data object model.
    
    N.B. This test would work as a standard TestCase.
    """
    
    def setUp(self):
        """Create test extract data."""
        super(TestExtract, self).setUp()
        #testnode
        run = Node('run')
        run_core = run.addnode('core')
        run_core.addnode('id', 'Run_Id')
        job = run.addnode('job')
        job_core = job.addnode('core')
        job_core.addnode('id', 1)
        job_core.addnode('name', 'Core_1')
        job_core.addnode('status', 'completed')
        job = run.addnode('job')
        job_core = job.addnode('core')
        job_core.addnode('id', 2)
        job_core.addnode('name', 'Core_2')
        job_core.addnode('status', 'failed')
        job = run.addnode('job')
        job_core = job.addnode('core')
        job_core.addnode('id', 3)
        job_core.addnode('name', 'Core_3')
        job_core.addnode('status', 'completed')
        self.testnode = run
        #testxml
        self.testxml = """<?xml version="1.0" ?><run><core><id>Run_Id</id></core><job><core><id>1</id><name>Core_1</name><status>completed</status></core></job><job><core><id>2</id><name>Core_2</name><status>failed</status></core></job><job><core><id>3</id><name>Core_3</name><status>completed</status></core></job></run>"""
        #testdom
        self.testdom = minidom.parseString(self.testxml)
    
    def tearDown(self):
        """Dereference test extract data."""
        self.testnode = None
        self.testxml = None
        self.testdom = None
        super(TestExtract, self).tearDown()
    
    def test_fromdom(self):
        """Test fromdom() creates a node from XML DOM."""
        expected = self.testnode
        actual = Node.fromdom(self.testdom)
        assert equal(expected, actual), 'node does not convert from dom as expected'

    def test_todom(self):
        """Test todom() writes a node to XML DOM."""
        expected = self.testdom.toxml()
        actual = self.testnode.todom().toxml()
        assert expected == actual, 'node does not convert to dom as expected'

    def test_fromxml(self):
        """Test fromxml() creates a node from XML string."""
        expected = self.testnode
        actual = Node.fromxml(self.testxml)
        assert equal(expected, actual), 'node does not convert from xml as expected'
        
    def test_toxml(self):
        """Test toxml() writes a node to XML string."""
        expected = self.testxml
        actual = self.testnode.toxml().replace('\n','')
        assert expected == actual, 'node does not convert to XML as expected'

    def test_toprettyxml(self):
        """Test toprettyxml() writes a node to pretty XML string."""
        expected = self.testdom.toprettyxml(indent='   ')
        actual = self.testnode.toprettyxml()
        assert expected == actual, 'node does not convert to pretty XML as expected'
        
    def test_constructor(self):
        """Test __init__() creates a node."""
        expected = Node.fromxml('<my-node/>')
        actual = Node('my-node')
        assert equal(expected, actual), 'created node is not as expected'
        
    def test_constructor_value(self):
        """Test __init__() creates a node with a value."""
        expected = Node.fromxml('<my-node>my-value</my-node>')
        actual = Node('my-node', 'my-value')
        assert equal(expected, actual), 'created node is not as expected'
    
    def test_addnode(self):
        """Test addnode() adds a node."""
        expected = Node.fromxml('<my-node><my-sub-node/></my-node>')
        actual = Node('my-node')
        actual.addnode('my-sub-node')
        assert equal(expected, actual), 'result of adding node is not as expected'
        
    def test_addnode_value(self):
        """Test addnode() adds a node with a value."""
        expected = Node.fromxml('<my-node><my-sub-node>my-value</my-sub-node></my-node>')
        actual = Node('my-node')
        actual.addnode('my-sub-node', 'my-value')
        assert equal(expected, actual), 'result of adding node is not as expected'

    def test_getnodes(self):
        """Test getnodes() returns list of current node."""
        expected = [self.testnode]
        actual = self.testnode.getnodes()
        assert str(expected) == str(actual), 'getnodes() does not return expected list of current node.'
        actual = self.testnode.getnodes('')
        assert equal(expected, actual), "getnodes('') does not return expected list of current node."

    def test_getnodes_name(self):
        """Test getnodes('job') returns list of job subnodes."""
        expected = [n for n in self.testnode.nodes if n.name == 'job'] #job subnodes
        actual = self.testnode.getnodes('job')
        assert equal(expected, actual), "getnodes('job') does not return expected list of job subnodes."

    def test_getnodes_wild(self):
        """Test getnodes('*') returns list of all subnodes."""
        expected = self.testnode.nodes #all subnodes
        actual = self.testnode.getnodes('*')
        assert equal(expected, actual), "getnodes('*') does not return expected list of all subnodes."

    def test_getnodes_path(self):
        """Test getnodes('job.core') returns list of core subnodes of job subnodes."""
        expected = []
        for jobnode in [n for n in self.testnode.nodes if n.name == 'job']: #job subnodes
            expected.extend([n for n in jobnode.nodes if n.name =='core']) #core subnodes
        actual = self.testnode.getnodes('job.core')
        assert equal(expected, actual), "getnodes('job.core') does not return expected list of core subnodes of job subnodes."

    def test_getnodes_path_wild(self):
        """Test getnodes('job.core.*') returns list of all subnodes of job.core subnodes."""
        expected = []
        for jobnode in [n for n in self.testnode.nodes if n.name == 'job']: #job subnodes
            for corenode in [n for n in jobnode.nodes if n.name =='core']: #core subnodes
                expected.extend(corenode.nodes) #all subnodes
        actual = self.testnode.getnodes('job.core.*')
        assert equal(expected, actual), "getnodes('job.core.*') does not return expected list of all subnodes of job.core subnodes."

    def test_getnodes_path_wild_path(self):
        """Test getnodes('job.*.id') returns list of id subnodes of job.* subnodes."""
        expected = []
        for jobnode in [n for n in self.testnode.nodes if n.name == 'job']: #job subnodes
            for subnode in jobnode.nodes: #all subnodes
                expected.extend([n for n in subnode.nodes if n.name == 'id']) #id subnodes
        actual = self.testnode.getnodes('job.*.id')
        assert equal(expected, actual), "getnodes('job.*.id') does not return expected list of id subnodes of job.* subnodes."

    def test_getnodes_wild_path(self):
        """Test getnodes('*.core.id') returns list of id subnodes of job.* subnodes."""
        expected = []
        for subnode in self.testnode.nodes: #all subnodes
            for corenode in [n for n in subnode.nodes if n.name =='core']: #core subnodes
                expected.extend([n for n in corenode.nodes if n.name == 'id']) #id subnodes
        actual = self.testnode.getnodes('*.core.id')
        assert equal(expected, actual), "getnodes('*.core.id') does not return expected list of id subnodes of *.core subnodes."

    def test_getnodes_wild_wild(self):
        """Test getnodes('*.*') returns list of all subnodes of all subnodes."""
        expected = []
        for subnode in self.testnode.nodes: #all subnodes
            expected.extend(subnode.nodes)#all subnodes
        actual = self.testnode.getnodes('*.*')
        assert equal(expected, actual), "getnodes('*.*') does not return expected list of all subnodes of all subnodes."

    def test_getnode(self):
        """Test getnode() returns first node of getnodes()."""
        expected = self.testnode.getnodes('job.core.id')[0]
        actual = self.testnode.getnode('job.core.id')
        assert equal(expected, actual), "getnode('job.core.id') does not return first job.core.id subnode as expected."
        
    def test_getnode_unknown(self):
        """Test getnode() returns None for unknown name."""
        expected = None
        actual = self.testnode.getnode('xxx')
        assert expected == actual, "getnode('xxx') does not return None."

    def test_getvalues(self):
        """Test getvalues() returns values of getnodes()."""
        expected = [n.value for n in self.testnode.getnodes('job.core.id')]
        actual = self.testnode.getvalues('job.core.id')
        assert expected == actual, "getvalues('job.core.id') does not return job.core.id subnode values as expected."

    def test_getvalue(self):
        """Test getvalue() returns first value of getvalues()."""
        expected = self.testnode.getvalues('job.core.id')[0]
        actual = self.testnode.getvalue('job.core.id')
        assert expected == actual, "getvalue('job.core.id') does not return first job.core.id subnode value as expected."

    def test_getvalue_unknown(self):
        """Test getvalue() returns None for unknown name."""
        expected = None
        actual = self.testnode.getvalue('xxx')
        assert expected == actual, "getvalue('xxx') does not return None."


        



from Ganga.GPIDev.Adapters.ISplitter import ISplitter
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem
from Ganga.Utility.Config import getConfig
from Ganga.Utility.files import expandfilename
from Ganga.GPIDev.Lib.Job import Job
import pickle
import os
import copy


class OptionsFileSplitter(ISplitter):

    '''Split a jobs based on a list of option file fragments

    This Splitter takes as argument a list of option file statements and will
    generate a job for each item in this list. The value of the indevidual list
    item will be appended to the master options file. A use case of this
    splitter would be to change a parameter in an algorithm (e.g. a cut) and to
    recreate a set of jobs with different cuts
    '''
    _name = "OptionsFileSplitter"
    docstr = "List of option-file strings, each list item creates a new subjob"
    _schema = Schema(Version(1, 0),
                     {'optsArray': SimpleItem(defvalue=[], doc=docstr)})

    def _create_subjob(self, job, inputdata):
        j = Job()
        j.copyFrom(job)
        j.splitter = None
        j.merger = None
        j.inputsandbox = []  # master added automatically
        j.inputfiles = []
        j.inputdata = inputdata

        return j

    def split(self, job):
        subjobs = []

        inputdata = job.inputdata
        if not job.inputdata:
            share_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),
                                      'shared',
                                      getConfig('Configuration')['user'],
                                      job.application.is_prepared.name,
                                      'inputdata',
                                      'options_data.pkl')

            if os.path.exists(share_path):
                f = open(share_path, 'r+b')
                inputdata = pickle.load(f)
                f.close()

        for i in self.optsArray:
            j = self._create_subjob(job, inputdata)
            j._splitter_data = i
            subjobs.append(j)
        return subjobs




# Change LogicalFile to be a wrapper for the DiracFile,
# make sure DiracFile replicates required functionality and add any
# possible translation layer here
from Ganga.Core import GangaException
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from GangaDirac.Lib.Files.DiracFile import DiracFile
from Ganga.Utility.logging import getLogger
logger = getLogger()


def strip_filename(name):
    if len(name) >= 4 and name[0:4].upper() == 'PFN:':
        msg = 'Can not create LogicalFile from string that begins w/ "PFN:".'\
              ' You probably want to create a MassStorageFile.'
        raise GangaException(msg)
    if len(name) >= 4 and name[0:4].upper() == 'LFN:':
        name = name[4:]
    return name


class LogicalFile(DiracFile):
    #  Logical File schema
    # Observing the 'old' 1.0 schema whilst preserving backwards compatability
    # with the fact that we're translating the object into a DiracFile in this
    # case
    _schema = Schema(Version(1, 0), {'name': SimpleItem(defvalue="", doc='the LFN filename a LogicalFile is constructed with'),
                                     'namePattern': SimpleItem(defvalue="", doc='pattern of the file name', transient=1),
                                     'localDir': SimpleItem(defvalue=None, copyable=1, typelist=['str', 'type(None)'],
                                                            doc='local dir where the file is stored, used from get and put methods', transient=1),
                                     'remoteDir': SimpleItem(defvalue="", doc='remote directory where the LFN is to be placed in the dirac base directory by the put method.', transient=1),
                                     'locations': SimpleItem(defvalue=[], copyable=1, typelist=['str'], sequence=1,
                                                             doc="list of SE locations where the outputfiles are uploaded", transient=1),
                                     'compressed': SimpleItem(defvalue=False, typelist=['bool'], protected=0,
                                                              doc='wheather the output file should be compressed before sending somewhere', transient=1),
                                     'lfn': SimpleItem(defvalue='', copyable=1, typelist=['str'],
                                                       doc='return the logical file name/set the logical file name to use if not '
                                                       'using wildcards in namePattern', transient=1),
                                     'guid': SimpleItem(defvalue='', copyable=1, typelist=['str'],
                                                        doc='return the GUID/set the GUID to use if not using wildcards in the namePattern.', transient=1),
                                     'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1, sequence=1, copyable=0,
                                                               typelist=[
                                         'GangaDirac.Lib.Files.DiracFile'],
        doc="collected files from the wildcard namePattern", transient=1),
        'failureReason': SimpleItem(defvalue="", protected=1, copyable=0, doc='reason for the upload failure', transient=1)
    })
    _name = "LogicalFile"

# TODO:  Add warning to User NOT to create these objects themselves and that they should
#       only be used for backwards compatability to load old jobs

    def __init__(self, name=""):

        super(LogicalFile, self).__init__(lfn=name)

        self.name = name

        logger.warning(
            "!!! LogicalFile has been deprecated, this is now just a wrapper to the DiracFile object")
        logger.warning(
            "!!! Please update your scripts before LogicalFile is removed")

        self._setLFNnamePattern(_lfn=self.name, _namePattern='')

    def __construct__(self, args):

        if len(args) >= 1:
            self.name = args[0]
            self._setLFNnamePattern(_lfn=self.name, _namePattern='')

        if (len(args) != 1) or (type(args[0]) is not type('')):
            super(LogicalFile, self).__construct__(args)
        else:
            self.name = strip_filename(args[0])

    def __setattr__(self, name, value):

        if name == "name":
            #elf.name = value
            self.lfn = value
            import os.path
            self.namePattern = os.path.basename(value)
            self.remoteDir = os.path.dirname(value)
        super(LogicalFile, self).__setattr__(name, value)

    def _attribute_filter__set__(self, attrib_name, value):
        if attrib_name == "name":
            self._setLFNnamePattern(lfn=value, namePattern='')
        return super(LogicalFile, self)._attribute_filter__set__(attrib_name, value)



from __future__ import absolute_import
##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: LCGSEFile.py,v 0.1 2011-02-12 15:40:00 idzhunov Exp $
##########################################################################

from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem

from Ganga.Utility.Config import getConfig
import Ganga.Utility.logging
from Ganga.GPIDev.Base.Proxy import GPIProxyObjectFactory
logger = Ganga.Utility.logging.getLogger()
from Ganga.Utility import GridShell

from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.GPIDev.Base.Proxy import getName

import re
import os
import copy

regex = re.compile('[*?\[\]]')


def getLCGConfig():
    return getConfig('Output')['LCGSEFile']['uploadOptions']

class LCGSEFile(IGangaFile):

    """LCGSEFile represents a class marking an output file to be written into LCG SE
    """

    _schema = Schema(Version(1, 1), {
        'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
        'localDir': SimpleItem(defvalue="", copyable=1, doc='local dir where the file is stored, used from get and put methods'),
        'joboutputdir': SimpleItem(defvalue="", doc='outputdir of the job with which the outputsandbox file object is associated'),
        'se': SimpleItem(defvalue=getLCGConfig()['dest_SRM'], copyable=1, doc='the LCG SE hostname'),
        'se_type': SimpleItem(defvalue='', copyable=1, doc='the LCG SE type'),
        'se_rpath': SimpleItem(defvalue='', copyable=1, doc='the relative path to the file from the VO directory on the SE'),
        'lfc_host': SimpleItem(defvalue=getLCGConfig()['LFC_HOST'], copyable=1, doc='the LCG LFC hostname'),
        'srm_token': SimpleItem(defvalue='', copyable=1, doc='the SRM space token, meaningful only when se_type is set to srmv2'),
        'SURL': SimpleItem(defvalue='', copyable=1, doc='the LCG SE SURL'),
        'port': SimpleItem(defvalue='', copyable=1, doc='the LCG SE port'),
        'locations': SimpleItem(defvalue=[], copyable=1, typelist=[str], sequence=1, doc="list of locations where the outputfiles were uploaded"),
        'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1, sequence=1, copyable=0, doc="collected files from the wildcard namePattern"),
        'failureReason': SimpleItem(defvalue="", protected=1, copyable=0, doc='reason for the upload failure'),
        'compressed': SimpleItem(defvalue=False, typelist=[bool], protected=0, doc='wheather the output file should be compressed before sending somewhere')})
    _category = 'gangafiles'
    _name = "LCGSEFile"
    _exportmethods = ["location", "setLocation", "get", "put", "getUploadCmd"]

    def __init__(self, namePattern='', localDir='', **kwds):
        """ namePattern is the pattern of the output file that has to be written into LCG SE
        """
        super(LCGSEFile, self).__init__()
        self.namePattern = namePattern
        self.localDir = localDir

        self.locations = []

        self.shell = GridShell.getShell()

    def __setattr__(self, attr, value):
        if attr == 'se_type' and value not in ['', 'srmv1', 'srmv2', 'se']:
            raise AttributeError('invalid se_type: %s' % value)
        super(LCGSEFile, self).__setattr__(attr, value)

    def __construct__(self, args):
        self.localDir = ''
        if len(args) == 1 and isinstance(args[0], str):
            self.namePattern = args[0]
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], str):
            self.namePattern = args[0]
            self.localDir = args[1]
        self.shell = GridShell.getShell()
        self.locations = []

    def _on_attribute__set__(self, obj_type, attrib_name):
        r = copy.deepcopy(self)
        if getName(obj_type) == 'Job' and attrib_name == 'outputfiles':
            r.locations = []
            r.localDir = ''
            r.failureReason = ''
        return r

    def __repr__(self):
        """Get the representation of the file."""

        return "LCGSEFile(namePattern='%s')" % self.namePattern

    def __get_unique_fname__(self):
        '''gets an unique filename'''

        import random
        import time

        uuid = (str(random.uniform(0, 100000000)) +
                '-' + str(time.time())).replace('.', '-')
        user = getConfig('Configuration')['user']

        fname = 'user.%s.%s' % (user, uuid)
        return fname

    def setLocation(self):
        """
        Sets the location of output files that were uploaded to lcg storage element from the WN
        """

        job = self.getJobObject()

        postprocessLocationsPath = os.path.join(
            job.outputdir, getConfig('Output')['PostProcessLocationsFileName'])
        if not os.path.exists(postprocessLocationsPath):
            return

        def lcgse_line_processor(line, lcgse_file):
            guid = line[line.find('->') + 2:]
            pattern = line.split(' ')[1]
            name = line.split(' ')[2].strip('.gz')

            if regex.search(lcgse_file.namePattern) is not None:
                d = LCGSEFile(namePattern=name)
                d.compressed = lcgse_file.compressed
                d.lfc_host = lcgse_file.lfc_host
                d.se = lcgse_file.se
                # todo copy also the other attributes
                lcgse_file.subfiles.append(GPIProxyObjectFactory(d))
                lcgse_line_processor(line, d)
            elif pattern == lcgse_file.namePattern:
                if guid.startswith('ERROR'):
                    logger.error("Failed to upload file to LCG SE")
                    logger.error(guid[6:])
                    lcgse_file.failureReason = guid[6:]
                    return
                lcgse_file.locations = guid

        for line in open(postprocessLocationsPath, 'r'):

            if line.strip() == '':
                continue

            if line.startswith('lcgse'):
                lcgse_line_processor(line.strip(), self)

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        return self.locations

    def getUploadCmd(self):

        vo = getConfig('LCG')['VirtualOrganisation']

        cmd = 'lcg-cr --vo %s ' % vo
        if self.se != '':
            cmd = cmd + ' -d %s' % self.se
        if self.se_type == 'srmv2' and self.srm_token != '':
            cmd = cmd + ' -D srmv2 -s %s' % self.srm_token

        # specify the physical location
        if self.se_rpath != '':
            cmd = cmd + \
                ' -P %s/ganga.%s/filename' % (self.se_rpath,
                                              self.__get_unique_fname__())

        return cmd

    def put(self):
        """
        Executes the internally created command for file upload to LCG SE, this method will
        be called on the client
        """
        import glob

        sourceDir = ''

        # if used as a stand alone object
        if self._getParent() is None:
            if self.localDir == '':
                logger.warning(
                    'localDir attribute is empty, don\'t know from which dir to take the file')
                return
            else:
                sourceDir = self.localDir
        else:
            job = self.getJobObject()
            sourceDir = job.outputdir
        import os
        os.environ['LFC_HOST'] = self.lfc_host

        fileName = self.namePattern

        if self.compressed:
            fileName = '%s.gz' % self.namePattern

        if regex.search(fileName) is not None:
            for currentFile in glob.glob(os.path.join(sourceDir, fileName)):
                cmd = self.getUploadCmd()
                cmd = cmd.replace('filename', currentFile)
                cmd = cmd + ' file:%s' % currentFile

                (exitcode, output, m) = self.shell.cmd1(
                    cmd, capture_stderr=True)

                d = LCGSEFile(namePattern=os.path.basename(currentFile))
                d.compressed = self.compressed
                d.lfc_host = self.lfc_host
                d.se = self.se
                # todo copy also the other attributes

                if exitcode == 0:

                    match = re.search('(guid:\S+)', output)
                    if match:
                        d.locations = output.strip()

                    # Alex removed this as more general approach in job.py after put() is called
                    # remove file from output dir if this object is attached to a job
                    # if self._getParent() is not None:
                    #    os.system('rm %s' % os.path.join(sourceDir, currentFile))

                else:
                    d.failureReason = output
                    if self._getParent() is not None:
                        logger.error("Job %s failed. One of the job.outputfiles couldn't be uploaded because of %s" % (
                            str(self._getParent().fqid), self.failureReason))
                    else:
                        logger.error(
                            "The file can't be uploaded because of %s" % (self.failureReason))

                self.subfiles.append(GPIProxyObjectFactory(d))

        else:
            logger.debug("sourceDir: %s" % sourceDir)
            logger.debug("fileName: %s" % fileName)
            currentFile = os.path.join(sourceDir, fileName)
            import os.path
            if os.path.isfile(currentFile):
                logger.debug("currentFile: %s exists!" % currentFile)
            else:
                logger.debug("currentFile: %s DOES NOT exist!" % currentFile)

            cmd = self.getUploadCmd()
            cmd = cmd.replace('filename', currentFile)
            cmd = cmd + ' file:%s' % currentFile

            logger.debug("cmd is: %s" % cmd)

            (exitcode, output, m) = self.shell.cmd1(cmd, capture_stderr=True)

            if exitcode == 0:

                match = re.search('(guid:\S+)', output)
                if match:
                    self.locations = output.strip()

                # Alex removed this as more general approach in job.py after put() is called
                # remove file from output dir if this object is attached to a job
                # if self._getParent() is not None:
                #    os.system('rm %s' % os.path.join(sourceDir, currentFile))

            else:
                self.failureReason = output
                if self._getParent() is not None:
                    logger.error("Job %s failed. One of the job.outputfiles couldn't be uploaded because of %s" % (
                        str(self._getParent().fqid), self.failureReason))
                else:
                    logger.error(
                        "The file can't be uploaded because of %s" % (self.failureReason))

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        lcgCommands = []

        for outputFile in outputFiles:
            lcgCommands.append('lcgse %s %s %s' % (
                outputFile.namePattern, outputFile.lfc_host,  outputFile.getUploadCmd()))
            logger.debug("OutputFile (%s) cmd for WN script is: %s" %
                         (outputFile.namePattern, outputFile.getUploadCmd()))

        import inspect
        script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                        'scripts/LCGSEFileWNScript.py')

        from Ganga.GPIDev.Lib.File import FileUtils
        script = FileUtils.loadScript(script_location, '###INDENT###')

        script = script.replace('###LCGCOMMANDS###', str(lcgCommands))
        script = script.replace('###PATTERNSTOZIP###', str(patternsToZip))
        script = script.replace('###INDENT###', indent)
        script = script.replace('###POSTPROCESSLOCATIONSFP###', postProcessLocationsFP)

        return script

    def get(self):
        """
        Retrieves locally all files matching this LCGSEFile object pattern
        """
        to_location = self.localDir

        if not os.path.isdir(self.localDir):
            if self._getParent() is not None:
                to_location = self.getJobObject().outputdir
            else:
                logger.info("%s is not a valid directory.... Please set the localDir attribute" % self.localDir)
                return

        # set lfc host
        os.environ['LFC_HOST'] = self.lfc_host

        vo = getConfig('LCG')['VirtualOrganisation']

        for location in self.locations:
            destFileName = os.path.join(to_location, self.namePattern)
            cmd = 'lcg-cp --vo %s %s file:%s' % (vo, location, destFileName)
            (exitcode, output, m) = self.shell.cmd1(cmd, capture_stderr=True)

            if exitcode != 0:
                logger.error(
                    'command %s failed to execute , reason for failure is %s' % (cmd, output))

    def getWNScriptDownloadCommand(self, indent):

        script = """\n

###INDENT###os.environ['LFC_HOST'] = '###LFC_HOST###'
###INDENT###cwDir = os.getcwd()
###INDENT###dwnCmd = 'lcg-cp --vo ###VO### lfn:/grid/###VO###/###LOCATION###/###NAMEPATTERN### file:%s' % os.path.join(cwDir, '###NAMEPATTERN###')
###INDENT###os.system(dwnCmd)
"""

        script = script.replace('###INDENT###', indent)
        script = script.replace('###LFC_HOST###', self.lfc_host)
        script = script.replace(
            '###VO###', getConfig('LCG')['VirtualOrganisation'])
        script = script.replace('###LOCATION###', self.se_rpath)
        script = script.replace('###NAMEPATTERN###', self.namePattern)

        return script

    def processWildcardMatches(self):
        if self.subfiles:
            return self.subfiles

        from fnmatch import fnmatch

        if regex.search(self.namePattern):
            # TODO namePattern shouldn't contain slashes and se_rpath should
            # not contain wildcards
            exitcode, output, m = self.shell.cmd1('lcg-ls lfn:/grid/' + getConfig(
                'LCG')['VirtualOrganisation'] + '/' + self.se_rpath, capture_stderr=True)

            for filename in output.split('\n'):
                if fnmatch(filename, self.namePattern):
                    subfile = LCGSEFile(namePattern=filename)
                    subfile.se_rpath = self.se_rpath
                    subfile.lfc_host = self.lfc_host

                    self.subfiles.append(GPIProxyObjectFactory(subfile))

# add LCGSEFile objects to the configuration scope (i.e. it will be
# possible to write instatiate LCGSEFile() objects via config file)
import Ganga.Utility.Config
Ganga.Utility.Config.config_scope['LCGSEFile'] = LCGSEFile



#!/usr/bin/env python
import Queue
import traceback
import collections
from Ganga.Core.GangaThread import GangaThread
from Ganga.Utility.execute import execute
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Config import getConfig
from Ganga.GPIDev.Base.Proxy import getName

from collections import namedtuple

logger = getLogger()
QueueElement = namedtuple('QueueElement',  ['priority', 'command_input', 'callback_func', 'fallback_func', 'name'])
CommandInput = namedtuple('CommandInput',  ['command', 'timeout', 'env', 'cwd', 'shell', 'python_setup', 'eval_includes', 'update_env'])
FunctionInput = namedtuple('FunctionInput', ['function', 'args', 'kwargs'])


class WorkerThreadPool(object):

    """
    Client class through which Ganga objects interact with the local DIRAC server.
    """
    __slots__ = ['__queue', '__worker_threads',
                 '_saved_num_worker', '_saved_thread_prefix', '_frozen', '_shutdown']

    def __init__(self, num_worker_threads=None, worker_thread_prefix='Worker_'):
        if num_worker_threads is None:
            num_worker_threads=getConfig('Queues')['NumWorkerThreads']
        self.__queue = Queue.PriorityQueue()
        self.__worker_threads = []

        self._saved_num_worker = num_worker_threads
        self._saved_thread_prefix = worker_thread_prefix

        self.__init_worker_threads(self._saved_num_worker, self._saved_thread_prefix)

        self._frozen = False
        self._shutdown = False

    def __init_worker_threads(self, num_worker_threads, worker_thread_prefix):
        if len(self.__worker_threads) > 0:
            logger.warning("Threads already started!")
            for i in self.__worker_threads:
                logger.info(
                    "Worker Thread: %s is already running!" % i.gangaName)
            return

        for i in range(num_worker_threads):
            t = GangaThread(name=worker_thread_prefix + str(i),
                            auto_register=False,
                            target=self.__worker_thread)
            t._Thread__args = (t,)
            t._name = worker_thread_prefix + str(i)
            t._command = 'idle'
            t._timeout = 'N/A'
            t.start()
            self.__worker_threads.append(t)

    def __worker_thread(self, thread):
        """
        Code run by worker threads to allow parallelism in Ganga.

        Can be used for executing non-blocking calls to local DIRAC server
        """
        # Occasionally when shutting down the Queue import at top has been garbage collected
        # and line "except Queue.Empty: continue" will throw
        # <type 'exceptions.AttributeError'>: 'NoneType' object has no attribute 'Empty'
        # im hoping that importing within the thread will avoid this.
        import Queue

        oldname = thread.gangaName

        # Note can use threading.current_thread to get the thread rather than passing it as an arg
        # easier to unit test this way though with a dummy thread.
        while not thread.should_stop():
            try:
                item = self.__queue.get(True, 0.05)
            except Queue.Empty:
                # wait 0.05 sec then loop again to give shutdown a chance
                continue

            # regster as a working thread
            if isinstance(item, QueueElement):
                oldname = thread.gangaName
                thread.gangaName = item.name

            thread.register()

            if not isinstance(item, QueueElement):
                logger.error("Unrecognised queue element: '%s'" % repr(item))
                logger.error("                  expected: 'QueueElement'")
                self.__queue.task_done()
                thread.unregister()
                continue

            if isinstance(item.command_input, FunctionInput):
                thread._command = getName(item.command_input.function)
            elif isinstance(item.command_input, CommandInput):
                thread._command = item.command_input.command
                thread._timeout = item.command_input.timeout
            else:
                logger.error("Unrecognised input command type: '%s'" % repr(item.command_input))
                logger.error("                       expected: ('FunctionInput' or 'CommandInput')")
                self.__queue.task_done()
                thread.unregister()
                continue

            try:
                if isinstance(item.command_input, FunctionInput):
                    these_args = item.command_input.args
                    if isinstance(these_args, str):
                        these_args = (these_args, )
                    result = item.command_input.function(*these_args, **item.command_input.kwargs)
                else:
                    result = execute(*item.command_input)
            except Exception as e:
                logger.error("Exception raised executing '%s' in Thread '%s':\n%s" % (thread._command, thread.gangaName, traceback.format_exc()))
                if item.fallback_func.function is not None:
                    if isinstance(item.fallback_func, FunctionInput):
                        thread._command = getName(item.fallback_func.function)
                        thread._timeout = 'N/A'
                        try:
                            item.fallback_func.function(e, *item.fallback_func.args, **item.fallback_func.kwargs)
                        except Exception as x:
                            logger.error("Exception raised in fallback function '%s' of Thread '%s':\n%s" % (thread._command, thread.gangaName, traceback.format_exc()))
                    else:
                        logger.error("Unrecognised fallback_func type: '%s'" % repr(item.fallback_func))
                        logger.error("                       expected: 'FunctionInput'")
            else:
                if item.callback_func.function is not None:
                    if isinstance(item.callback_func, FunctionInput):
                        thread._command = getName(item.callback_func.function)
                        thread._timeout = 'N/A'
                        try:
                            item.callback_func.function(
                                result, *item.callback_func.args, **item.callback_func.kwargs)
                        except Exception as e:
                            logger.error("Exception raised in callback_func '%s' of Thread '%s': %s" % (
                                thread._command, thread.gangaName, traceback.format_exc()))
                    else:
                        logger.error("Unrecognised callback_func type: '%s'" % repr(item.callback_func))
                        logger.error("                       expected: 'FunctionInput'")
            finally:
                # unregister as a working thread bcoz free
                thread._command = 'idle'
                thread._timeout = 'N/A'
                self.__queue.task_done()
                thread.unregister()

            thread.gangaName = oldname

    def add_function(self,
                     function, args=(), kwargs={}, priority=5,
                     callback_func=None, callback_args=(), callback_kwargs={},
                     fallback_func=None, fallback_args=(), fallback_kwargs={},
                     name=None):

        if not isinstance(function, collections.Callable):
            logger.error('Only a python callable object may be added to the queue using the add_function() method')
            return
        if self._frozen is True:
            if not self._shutdown:
                logger.warning("Cannot Add Process as Queue is frozen!")
            return
        self.__queue.put(QueueElement(priority=priority,
                                      command_input=FunctionInput(
                                          function, args, kwargs),
                                      callback_func=FunctionInput(
                                          callback_func, callback_args, callback_kwargs),
                                      fallback_func=FunctionInput(fallback_func, fallback_args, fallback_kwargs), name=name
                                      ))

    def add_process(self,
                    command, timeout=None, env=None, cwd=None, shell=False,
                    python_setup='', eval_includes=None, update_env=False, priority=5,
                    callback_func=None, callback_args=(), callback_kwargs={},
                    fallback_func=None, fallback_args=(), fallback_kwargs={},
                    name=None):

        if not isinstance(command, str):
            logger.error("Input command must be of type 'string'")
            return
        if self._frozen is True:
            if self._shutdown:
                logger.warning("Cannot Add Process as Queue is frozen!")
            return
        self.__queue.put(QueueElement(priority=priority,
                                      command_input=CommandInput(
                                          command, timeout, env, cwd, shell, python_setup, eval_includes, update_env),
                                      callback_func=FunctionInput(
                                          callback_func, callback_args, callback_kwargs),
                                      fallback_func=FunctionInput(fallback_func, fallback_args, fallback_kwargs), name=name
                                      ))

    def map(self, function, *iterables):
        if not isinstance(function, collections.Callable):
            raise TypeError('must be a function')
        if self._frozen is True:
            logger.error("Cannot map a Function as Queue is frozen!")
        for args in zip(*iterables):
            self.__queue.put(QueueElement(priority=5,
                                          command_input=FunctionInput(
                                              function, args, {})
                                          ))

    def clear_queue(self):
        """
        Purges the thread pools queue.
        """
        self.__queue.queue = []

    def get_queue(self):
        """
        Returns the current state of the multiprocess queue that the local DIRAC server is working through.
        """
        return self.__queue.queue[:]

    def worker_status(self):
        """
        Returns a informatative tuple containing the threads name, current command it's working on and the timeout for that command.
        """
        return [(w.gangaName, w._command, w._timeout) for w in self.__worker_threads]

    def _stop_worker_threads(self, shutdown=False):
        self._shutdown = shutdown
        for w in self.__worker_threads:
            w.stop()
            w.join()
            # FIXME NEED TO CALL AN OPTIONAL CLEANUP FUCNTION HERE IF THREAD IS STOPPED
            # w.unregister()
            #del w
        self.__worker_threads = []
        return

    def _start_worker_threads(self):
        if len(self.__worker_threads) > 0:
            self._stop_worker_threads()

        self.__init_worker_threads(
            self._saved_num_worker, self._saved_thread_prefix)
        return

###################################################################




"""Robot Framework.

The Driver module contains a Driver class which executes a run and a
loaddriver() factory method for creating driver instances based on the Robot
configuration.

The Action module contains an IAction interface that any action to be executed
by Driver should implement.

The Utility module contains utility functions used throughout the robot code.

"""


##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: ShutdownManager.py,v 1.1 2008-07-17 16:40:50 moscicki Exp $
#
# Copyright (C) 2003-2007 The Ganga Project
#
# This file is part of Ganga.
#
# Ganga is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Ganga is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
##########################################################################
"""
Extend the behaviour of the default *atexit* module to support:

 1) automatically catching of the (possible) exceptions thrown by exit function
 
 2) exit function prioritization (lower value means higher priority) 
  E.g:
    import atexit
    atexit.register((<PRIORITY>,myfunc),args)  
   
  The backward-compatibility is kept so the existing code using :
    import atexit
    atexit.register(myfunc,args)
  registers the function with the lowest priority (sys.maxint)
"""

import atexit

from Ganga.Utility import stacktracer


def _ganga_run_exitfuncs():
    """run any registered exit functions

    atexit._exithandlers is traversed based on the priority.
    If no priority was registered for a given function
    than the lowest priority is assumed (LIFO policy)

    We keep the same functionality as in *atexit* bare module but
    we run each exit handler inside a try..catch block to be sure all
    the registered handlers are executed
    """

    from Ganga.GPIDev.Base.Proxy import getName

    #print("Shutting Down Ganga Repositories")
    from Ganga.Runtime import Repository_runtime
    Repository_runtime.flush_all()

    from Ganga.Utility.logging import getLogger
    logger = getLogger()

    # Set the disk timeout to 1 sec, sacrifice stability for quick-er exit
    from Ganga.Utility.Config import setConfigOption
    setConfigOption('Configuration', 'DiskIOTimeout', 1)

    ## Stop the Mon loop from iterating further!
    from Ganga.Core import monitoring_component
    if monitoring_component is not None:
        from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import getStackTrace
        getStackTrace()
        monitoring_component.disableMonitoring()
        monitoring_component.stop()
        monitoring_component.join()

    ## Stop the tasks system from running it's GangaThread before we get to the GangaThread shutdown section!
    from Ganga.GPIDev.Lib.Tasks import stopTasks
    stopTasks()

    # Set the disk timeout to 3 sec, sacrifice stability for quick-er exit
    #from Ganga.Utility.Config import setConfigOption
    #setConfigOption('Configuration', 'DiskIOTimeout', 3)

    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import _purge_actions_queue, stop_and_free_thread_pool
    _purge_actions_queue()
    stop_and_free_thread_pool()

    def priority_cmp(f1, f2):
        """
        Sort the exit functions based on priority in reversed order
        """
        # extract the priority number from the function element
        p1 = f1[0][0]
        p2 = f2[0][0]
        # sort in reversed order
        return cmp(p2, p1)

    def add_priority(x):
        """
        add a default priority to the functions not defining one (default priority=sys.maxint)
        return a list containg ((priority,func),*targs,*kargs) elements
        """
        import sys
        func = x[0]
        if isinstance(func, tuple) and len(x[0]) == 2:
            return x
        else:
            new = [(sys.maxsize, func)]
            new.extend(x[1:])
            return new

    atexit._exithandlers = map(add_priority, atexit._exithandlers)
    atexit._exithandlers.sort(priority_cmp)

    logger.info("Stopping Job processing before shutting down Repositories")

    from Ganga.Core.GangaThread.WorkerThreads import _global_queues
    if _global_queues:
        _global_queues.freeze()

    import inspect
    while atexit._exithandlers:

        (priority, func), targs, kargs = atexit._exithandlers.pop()
        try:
            if hasattr(func, 'im_class'):
                for cls in inspect.getmro(func.__self__.__class__):
                    if getName(func) in cls.__dict__:
                        logger.debug(getName(cls) + " : " + getName(func))
                func(*targs, **kargs)
            else:
                logger.debug("noclass : " + getName(func))
                #print("%s" % str(func))
                #print("%s" % str(inspect.getsourcefile(func)))
                #func(*targs, **kargs)
                ## This attempts to check for and remove externally defined shutdown functions which may interfere with the next steps!
                if str(inspect.getsourcefile(func)).find('Ganga') != -1:
                    func(*targs, **kargs)
        except Exception as err:
            s = 'Cannot run one of the exit handlers: %s ... Cause: %s' % (getName(func), str(err))
            logger.debug(s)
            try:
                import os
                logger.debug("%s" % os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(func)))) )
                logger.debug("\n%s" % inspect.getsource(func))
            except Exception as err2:
                logger.debug("Error getting source code and failure reason: %s" % str(err2))

    from Ganga.Core.GangaThread.WorkerThreads import shutDownQueues
    shutDownQueues()

    logger.info("Shutting Down Ganga Repositories")
    from Ganga.Runtime import Repository_runtime
    Repository_runtime.shutdown()

    from Ganga.Core.InternalServices import Coordinator
    Coordinator.servicesEnabled = False

    from Ganga.Core.GangaRepository.SessionLock import removeGlobalSessionFiles, removeGlobalSessionFileHandlers
    removeGlobalSessionFileHandlers()
    removeGlobalSessionFiles()

    if stacktracer._tracer:
        stacktracer.trace_stop()

    from Ganga.Utility.logging import requires_shutdown, final_shutdown
    if requires_shutdown is True:
        final_shutdown()

    from Ganga.Runtime import bootstrap
    if bootstrap.DEBUGFILES or bootstrap.MONITOR_FILES:
        bootstrap.printOpenFiles()


def install():
    """
    Install a new shutdown manager, by overriding methods from atexit module
    """
    # override the atexit exit function
    atexit._run_exitfuncs = _ganga_run_exitfuncs
    #del atexit

    # override the default exit function
    import sys
    sys.exitfunc = atexit._run_exitfuncs

#
#$Log: not supported by cvs2svn $
# Revision 1.2  2007/07/27 14:31:56  moscicki
# credential and clean shutdown updates from Adrian (from Ganga-4-4-0-dev-branch)
#
# Revision 1.1.2.2  2007/07/27 13:03:17  amuraru
#*** empty log message ***



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Shell.py,v 1.7 2009-06-26 11:35:09 moscicki Exp $
##########################################################################
#
# Shell wrapper with environment caching
#
# Usage:
#
#
# Initialisation: The shell script is sourced and the environment is captured
#
#     shell = Shell('/afs/cern.ch/project/gd/LCG-share/sl3/etc/profile.d/grid_env.sh')
#
# Output is returned in a file
#
#     rc,outfile,m=shell.cmd('edg-get-job-status -all')
#
# Output is returned as a string
#
#     rc,output,m=shell.cmd1('edg-get-job-status -all')
#
# Output is not captured. Useful for commands that require interactions
#
#     rc=shell.system('grid-proxy-init')
#
# Wrapper script is written for command including the setting of the
# environment first. Useful for situations where it is an external Python
# module that is calling command. It is callers responsibility to enter
# new location into PATH as this might have external effects.
#
#     fullpath=shell.wrapper('lcg-cp')

import errno
import os
import re
import stat
import tempfile
import time
import signal
import subprocess
import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()


def expand_vars(env):
    tmp_dict = {}
    for k, v in env.iteritems():
        if not str(v).startswith('() {'):
            if not str(k).endswith('()'):
                tmp_dict[k] = os.path.expandvars(v)
        # Be careful with exported bash functions!
        elif not str(k).endswith('()'):
            this_string = str(v).split('\n')
            final_str = ""
            for line in this_string:
                final_str += str(os.path.expandvars(line)).strip()
                if not final_str.endswith(';'):
                    final_str += " ;"
                final_str += " "
            tmp_dict[k] = final_str
            # print tmp_dict[k]
    return tmp_dict


class Shell(object):

    def __init__(self, setup=None, setup_args=[]):
        """

        THIS EXPECTS THE BASH SHELL TO AT LEAST BE AVAILABLE TO RUN THESE COMMANDS!

        The setup script is sourced (with possible arguments) and the
        environment is captured. The environment variables are expanded
        automatically (this is a fix for bug #44259: GangaLHCb tests fail due to
        gridProxy check).

        Example of variable expansion:

        os.environ['BAR'] = 'rabarbar'
        os.environ['FOO'] = '$BAR'
        s = Shell() # with or without the setup script
        assert s.env['FOO'] == 'rabarbar' # NOT literal string '$BAR'

        NOTE: the behaviour is not 100% bash compatible: undefined variables in
        bash are empty strings, Shell() leaves the literals unchanged,so:

        os.environ['FOO'] = '$NO_BAR'
        s = Shell()
        if 'NO_BAR' not in os.environ:
           assert s.env['FOO'] == '$NO_BAR'

        """

        if setup is not None:

            env = dict(os.environ)
            env = expand_vars(env)

            logger.debug("Initializing Shell")
            logger.debug("command:\n%s" % setup)
            logger.debug("args:\n%s" % " ".join(setup_args))

            this_cwd = os.path.abspath(os.getcwd())
            if not os.path.exists(this_cwd):
                this_cwd = os.path.abspath(tempfile.gettempdir())
            logger.debug("Using CWD: %s" % this_cwd)
            command = 'source %s %s > /dev/null 2>&1; python -c "from __future__ import print_function; import os; print(os.environ)"' % (setup, " ".join(setup_args))
            logger.debug('Running:   %s' % command )
            pipe = subprocess.Popen('bash', env=env, cwd=this_cwd, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            output = pipe.communicate(command)
            rc = pipe.poll()
            if rc:
                logger.warning('Unexpected rc %d from setup command %s', rc, setup)

            try:
                env2 = expand_vars(eval(eval(str(output))[0]))
            except Exception as err:
                logger.debug("Err: %s" % str(err))
                env2 = None
                logger.error("Cannot construct environ:\n%s" % str(output))
                try:
                    logger.error("eval: %s" % str(eval(str(output)[0])))
                except Exception as err2:
                    logger.debug("Err2: %s" % str(err2))
                    pass
                try:
                    logger.error("eval(eval): %s" % eval(eval(str(output))[0]))
                except Exception as err3:
                    logger.debug("Err3: %s" % str(err3))
                    pass

            if env2:
                env = env2

            self.env = env
        else:
            # bug #44334: Ganga/Utility/Shell.py does not save environ
            env = dict(os.environ)
            self.env = expand_vars(env)

        self.dirname = None

    def pythonCmd(self, cmd, soutfile=None, allowed_exit=None,
                  capture_stderr=False, timeout=None, mention_outputfile_on_errors=True):
        "Execute a python command and captures the stderr and stdout which are returned in a file"

        if allowed_exit is None:
            allowed_exit = [0]
        return self.cmd(cmd, soutfile, allowed_exit, capture_stderr, timeout, mention_outputfile_on_errors, python=True)

    def cmd(self, cmd, soutfile=None, allowed_exit=None,
            capture_stderr=False, timeout=None, mention_outputfile_on_errors=True, python=False):
        "Execute an OS command and captures the stderr and stdout which are returned in a file"

        if allowed_exit is None:
            allowed_exit = [0]

        if not soutfile:
            soutfile = tempfile.NamedTemporaryFile(mode='w+t', suffix='.out', delete=False).name

        logger.debug('Running shell command: %s' % cmd)
        try:
            t0 = time.time()
            already_killed = False
            timeout0 = timeout
            launcher = ['/bin/sh', '-c']
            args = ['%s > %s 2>&1' % (cmd, soutfile)]
            command = launcher
            for i in args:
                command.append(i)
            this_cwd = os.path.abspath(os.getcwd())
            if not os.path.exists(this_cwd):
                this_cwd = os.path.abspath(tempfile.gettempdir())
            logger.debug("Using CWD: %s" % this_cwd)

            #import traceback
            # traceback.print_stack()

            process = subprocess.Popen(command, env=self.env, cwd=this_cwd)
            pid = process.pid
            while True:
                wpid, sts = os.waitpid(pid, os.WNOHANG)
                if wpid != 0:
                    if os.WIFSIGNALED(sts):
                        rc = -os.WTERMSIG(sts)
                        break
                    elif os.WIFEXITED(sts):
                        rc = os.WEXITSTATUS(sts)
                        break
                if timeout and time.time() - t0 > timeout:
                    logger.warning(
                        'Command interrupted - timeout %ss reached: %s', timeout0, cmd)
                    if already_killed:
                        sig = signal.SIGKILL
                    else:
                        sig = signal.SIGTERM
                    logger.debug('killing process %d with signal %d', pid, sig)
                    os.kill(pid, sig)
                    t0 = time.time()
                    # wait just 5 seconds before killing with SIGKILL
                    timeout = 5
                    already_killed = True
                time.sleep(0.05)

        except OSError as e:
            if e.errno == 10:
                rc = process.returncode
                logger.debug("Process has already exitted which will throw a 10")
                logger.debug("Exit status is: %s" % rc)
            else:
                logger.warning('Problem with shell command: %s, %s', e.errno, e.strerror)
                rc = 255

        BYTES = 4096
        if rc not in allowed_exit:
            logger.warning('exit status [%d] of command %s', rc, cmd)
            if mention_outputfile_on_errors:
                logger.warning('full output is in file: %s', soutfile)
            with open(soutfile) as sout_file:
                logger.warning('<first %d bytes of output>\n%s', BYTES, sout_file.read(BYTES))
            logger.warning('<end of first %d bytes of output>', BYTES)

        # FIXME /bin/sh might have also other error messages
        m = None
        if rc != 0:
            with open(soutfile) as sout_file:
                m = re.search('command not found\n', sout_file.read())
            if m:
                logger.warning('command %s not found', cmd)

        return rc, soutfile, m is None

    def cmd1(self, cmd, allowed_exit=None, capture_stderr=False, timeout=None, python=False):
        "Executes an OS command and captures the stderr and stdout which are returned as a string"

        if allowed_exit is None:
            allowed_exit = [0]

        rc, outfile, m = self.cmd(cmd, None, allowed_exit, capture_stderr,
                                  timeout, mention_outputfile_on_errors=False, python=python)

        from contextlib import closing
        with closing(open(outfile)) as out_file:
            output = out_file.read()
        try:
            os.remove(outfile)
        except OSError as err:
            if err.errno != errno.ENOENT:
                logger.debug("Err removing shell output: %s" % str(err))
                raise err
            pass

        return rc, output, m

    def system(self, cmd, allowed_exit=None, stderr_file=None):
        """Execute on OS command. Useful for interactive commands. Stdout and Stderr are not
        caputured and are passed on the caller.

        stderr_capture may specify a name of a file to which stderr is redirected.
        """
        if allowed_exit is None:
            allowed_exit = [0]

        logger.debug('Calling shell command: %s' % cmd)

        if stderr_file:
            cmd += " 2> %s" % stderr_file

        try:
            rc = subprocess.call(['/bin/sh', '-c', cmd], env=self.env)
        except OSError as e:
            logger.warning(
                'Problem with shell command: %s, %s', e.errno, e.strerror)
            rc = 255
        return rc

    def wrapper(self, cmd, preexecute=None):
        """Write wrapper script for command

        A wrapper around cmd is written including the setting of the environment.
        Useful for situations where it is an external Python module that is
        calling the command. It is callers responsibility to enter
        new location into PATH as this might have external effects. Full path of
        wrapper script is returned. Preexecute can contain extra commands to be
        executed before cmd

        fullpath = s.wrapper('lcg-cp', 'echo lcg-cp called with arguments $*'"""

        if not self.dirname:
            self.dirname = tempfile.mkdtemp()

        fullpath = os.path.join(self.dirname, cmd)
        with open(fullpath, 'w') as f:
            f.write("#!/bin/bash\n")
            for k, v in self.env.iteritems():
                f.write("export %s='%s'\n" % (k, v))
            if preexecute:
                f.write("%s\n" % preexecute)
            f.write("%s $*\n" % cmd)
        os.chmod(fullpath, stat.S_IRWXU)

        return fullpath

#
#
# $Log: not supported by cvs2svn $
# Revision 1.6  2008/11/27 15:06:08  moscicki
# bug #44393: wrong redirection of stdout/stderr in Shell
#
# Revision 1.5  2008/11/21 15:42:03  moscicki
# bug #44259: GangaLHCb tests fail due to gridProxy check
#
# Revision 1.4  2008/11/21 14:03:36  moscicki
# bug #44334: Ganga/Utility/Shell.py does not save environ
#
# Revision 1.3  2008/11/07 12:26:12  moscicki
# fixed Shell in case the setup script produces garbage on stdout (now discarded to /dev/null)
#
# Revision 1.2  2008/10/24 06:42:14  moscicki
# bugfix #40932: Ganga incompatible with shell functions (using os.environ directly instead of printenv)
#
# Revision 1.1  2008/07/17 16:41:00  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.5.12.4  2008/07/02 13:18:54  moscicki
# syntax error fix
#
# Revision 1.5.12.3  2008/07/01 14:45:03  moscicki
# fixes to support ARC (Nordu Grid)
#
# Revision 1.5.12.2  2008/02/21 12:09:41  amuraru
# bugfix #33685
#
# Revision 1.5.12.1  2007/12/10 19:19:59  amuraru
# merged changes from Ganga 4.4.4
#
# Revision 1.7  2007/10/23 14:45:58  amuraru
# fixed a bug in wrapper function to pass the command line arguments
#
# Revision 1.6  2007/10/15 14:16:57  amuraru
# [ulrik[ Added wrapper method to allow constructing a  wrapper around the command to
# include the setting of the environment. Useful for situations where it is an external
# Python module that is calling the command.
#
# Revision 1.5  2007/06/08 07:57:12  moscicki
# stderr capture option for Shell.system()
#
# Revision 1.4  2005/12/15 12:20:09  moscicki
# fix from uegede: OSError handling to avoid the waitpid problem
#
# Revision 1.3  2005/09/21 09:05:01  andrew
# removed unecessary redirection lines from the "system" method. Since the
# purpose of the system method is to be interactive, it does not make sense
# to capture std[in|out|err]
#
#
#



from Ganga.Core.exceptions import ApplicationConfigurationError
from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *

from Ganga.Utility.Config import makeConfig, ConfigError
from Ganga.Utility.logging import getLogger
from Ganga.Utility.files import expandfilename

from Ganga.GPIDev.Adapters.IApplication import IApplication
from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

from Ganga.Lib.Mergers.Merger import *

import urlparse

# ------------------------------------------------------
# Main Spider application class
class Spider(IApplication):
    """The main Spider executable"""

    _schema = Schema(Version(2,0), {
        'domains'       : SimpleItem(defvalue=[''],doc='The domains to spider'),
        'safe_domains'  : SimpleItem(defvalue=[''],doc='Safe domains to spider'),
        'image_ext'     : SimpleItem(defvalue=[''],doc='Image Extensions to record'),
        'file_ext'      : SimpleItem(defvalue=[''],doc='File Extensions to record'),
        'repository_loc': SimpleItem(defvalue='', doc = 'The location of the spider repository'),
        'max_links'     : SimpleItem(defvalue=1000, doc = 'Maximum links per domain'),
        'payload'       : FileItem(doc = 'Payload file to use'),
        'payload_output': SimpleItem( defvalue = '', doc='payload output')
        })
    _category = 'applications'
    _name = 'Spider'
    _exportmethods = []
    _queuelistfilename = 'queue_list.txt'
    _inputlistfilename = 'input_list.txt'
    _extensionfilename = 'ext.txt'
    _safedomsfilename = 'safe_doms.txt'
    
    def configure(self,masterappconfig):
        logger.debug('Spider configure called')
        return (None,None)
    
    def master_configure(self):
        logger.debug('Spider master_configure called')
        
        # check domains have been asked for
        if len(self.domains) == 0:
            raise ApplicationConfigurationError(None,'No domains specified')

        # check the payload
        if self.payload.name != "":
            payload = __import__(os.path.basename(self.payload.name.strip('.py')))
            payload.__getattribute__('init')
            payload.__getattribute__('new_html')
            payload.__getattribute__('end')

        return (0, None)

    def postprocess(self):
        """Sort out the results of the spider"""
        from Ganga.GPIDev.Lib.Job import Job
        j = self.getJobObject()

        # extract all the spidered links and images, etc from the job
        viewed_file = open( os.path.join(j.outputdir, 'viewed_list.txt'), 'r')
        viewed_list = viewed_file.readlines()
        viewed_file.close()
        self.ExtractViewedLinks( viewed_list )
            
        queued_file = open( os.path.join(j.outputdir, 'links_list.txt'), 'r')
        queued_list = queued_file.readlines()
        queued_file.close()
        self.ExtractQueuedLinks( queued_list )
            
        dom_list_file = open( os.path.join(j.outputdir, 'dom_list.txt'), 'r')
        dom_list = dom_list_file.readlines()
        dom_list_file.close()
        if len(dom_list) != 0:
            self.ExtractNewDomains( dom_list )
                        
        image_list_file = open( os.path.join(j.outputdir, 'img_list.txt'), 'r')
        image_info_file = open( os.path.join(j.outputdir, 'image_info.txt'), 'r')
        self.ExtractInfo(image_list_file, image_info_file, 'ImageInfo.txt')
        image_list_file.close()
        image_info_file.close()

        ext_list_file = open( os.path.join(j.outputdir, 'ext_list.txt'), 'r')
        ext_info_file = open( os.path.join(j.outputdir, 'ext_info.txt'), 'r')
        self.ExtractInfo(ext_list_file, ext_info_file, 'ExtInfo.txt')
        ext_list_file.close()
        ext_info_file.close()

    def ExtractViewedLinks(self, viewed_list ):

        # first, get together all similar domains
        viewed_arr = {}
        for ln in viewed_list:
            
            url = urlparse.urlparse( ln )
            if not url[1] in viewed_arr.keys():
                viewed_arr[url[1]] = []
            viewed_arr[url[1]].append( ln )

        # now add these to the lists in the directories
        for list_key in viewed_arr:

            dom = os.path.join( self.repository_loc, urlparse.urlparse( viewed_arr[list_key][0].strip() )[1] )
        
            if not os.path.exists( dom ):
                print "Error in ExtractViewedResults: Could not find domain " + dom
                continue
        
            viewed_file = open( os.path.join(dom, 'viewed_list.txt'), 'a')
        
            for ln in  viewed_arr[list_key]:
                viewed_file.write(ln)

            viewed_file.close()

        return

    def ExtractQueuedLinks( self, queued_list ):

        import sys
        
        # first, get together all similar domains
        queued_arr = { }
        temp_list = { }
    
        for ln in queued_list:
        
            url = urlparse.urlparse( ln )
            if not url[1] in temp_list.keys():
                temp_list[url[1]] = [ ]
            temp_list[url[1]].append( ln.strip('\n') )

        # create the sets
        for list_key in temp_list:
            queued_arr[list_key] = set( temp_list[list_key] )
            
        # now add these to the lists in the directories
        for list_key in temp_list:

            dom = os.path.join( self.repository_loc, urlparse.urlparse( temp_list[list_key][0].strip() )[1])
            if not os.path.exists( dom ):
                print "Error in ExtractQueuedResults: Could not find domain " + dom + " (" + temp_list[list_key][0].strip() + ")"
                continue

            queued_file = open( os.path.join(dom, self._queuelistfilename), 'r')
            temp_list2 = [ ]
            for ln in queued_file.readlines():
                temp_list2.append( ln.strip('\n') )
            queued_file.close()
            old_queue_list = set( temp_list2 )
         
            viewed_file = open( os.path.join(dom, 'viewed_list.txt'), 'r')

            temp_list2 = []
            for ln in viewed_file.readlines():
                temp_list2.append( ln.strip('\n') )
                
            viewed_list = set( temp_list2 )
            
            # find the links that aren't in the viewed list and is the union of the queued list
            new_queued_list = (queued_arr[list_key] - viewed_list) | old_queue_list

            queued_file = open( os.path.join(dom, self._queuelistfilename), 'w')
            for ln in new_queued_list:
                queued_file.write(ln + '\n')
            queued_file.close()

        return

    def ExtractNewDomains( self, dom_list ):

        # extract any new domains and add them to the domain list
        import re
        nonalpha = re.compile('[^a-zA-Z.]')
        for dom in dom_list:
            if not nonalpha.search(dom):
                self.domains.append(dom.strip('\n'))
    
        return

    def ExtractInfo( self, image_list_file, image_info_file, output_file ):
    
        # go through image list
        info_list = image_info_file.readlines()
        num = 0
    
        for ln in image_list_file.readlines():
        
            url = urlparse.urlparse( ln )
            dir = os.path.join( self.repository_loc, url[1] )
            if os.path.exists( dir ):
                try: 
                    image_file = open( os.path.join(dir, output_file), 'r')
                    image_list = image_file.readlines()
                    image_file.close()
                except:
                    image_file = open( os.path.join(dir, output_file), 'w')
                    image_list = []
                    image_file.close()
                
                if not ln in image_list:
                    #open( os.path.join(dir, output_file), 'a').write( ln )

                    # now find the entry in the image_info.txt file
                    i = 0
                    for im in info_list:
                        if im.find(ln.strip()) != -1:
                            out_info = open( os.path.join(dir, output_file), "a")
                            out_info.write( info_list[i-1] )
                            out_info.write( info_list[i] )
                            out_info.write( info_list[i+1] )
                            out_info.write( info_list[i+2] )
                            out_info.write( info_list[i+3] )
                            out_info.write( info_list[i+4] )
                            out_info.close()
                            
                            break
                    
                        i=i+1



# Copyright information
__author__  = "Ole Weidner <oweidner@cct.lsu.edu>"
__date__    = "16 September 2010"
__version__ = "1.0"

##############################################################################
##
class SAGAWrapperScript:

    _attributes = ('wrapper_script_template', 'exe', 'args', 'envi', 'sandbox', 'imods', 'mons', 'out')

    ##########################################################################
    ##    
    def setExecutable(self, exe):
        self.exe = exe

    ##########################################################################
    ##    
    def setOutputPatterns(self, out):
        self.out = out

    ##########################################################################
    ##    
    def setArguments(self, args):
        self.args = args
        
    ##########################################################################
    ##    
    def setEnvironment(self, envi):
        self.envi = envi
        
    ##########################################################################
    ##    
    def setInputSandbox(self, sandbox):
        self.sandbox = sandbox
        
    ##########################################################################
    ##    
    def setInlineModules(self, imods):
        self.imods = imods
      
    ##########################################################################
    ##
    def getScript(self):
        script = self.wrapper_script_template
        script = script.replace('###INLINEMODULES###', self.imods)
        script = script.replace('###EXECUTABLE###', repr(self.exe))
        script = script.replace('###ARGUEMTNS_AS_LIST###', repr(self.args))
        #script = script.replace('###ENVIRONMENT_AS_DICT###', repr(self.envi))
        script = script.replace('###INPUTSANDBOXFILE###', repr(self.sandbox))
        script = script.replace('###OUTPUTPATTERNS###', repr(self.out))
    
        return script

    ##########################################################################
    ##
    def __init__(self):
        self.imods = ''
        
        self.wrapper_script_template  = """#!/usr/bin/env python

import shutil
import os
import time
import sys

###INLINEMODULES###

executable       = ###EXECUTABLE###
arguments        = ###ARGUEMTNS_AS_LIST###
outputpatterns   = ###OUTPUTPATTERNS###


inputsandboxfile = ###INPUTSANDBOXFILE###

#redirect stdout/stderr to file
sys.stdout=file('./__syslog__','w')
sys.stderr=sys.stdout


## First things first. Unpack the input sandbox, since that's where
## all our files are. 
import tarfile

if os.path.exists(inputsandboxfile): 
    tar = tarfile.open(inputsandboxfile)
    tar.extractall()
    tar.close()

## Try to import the subprocess library. If it's not in the 
## PYTHON_PATH, we use the one pre-staged with this job. 
##
try:
    import subprocess
except ImportError as x:
    sys.path.insert(0, PYTHON_DIR)
    import subprocess
    


#sysout2 = os.dup(sys.stdout.fileno())
#syserr2 = os.dup(sys.stderr.fileno())

#sys.stdout.flush()
#sys.stderr.flush()

sys.path.insert(0, PYTHON_DIR)


result = 255

## EXECUTE THE STUFF WE WANT TO RUN
##

outfile = open('out.log', 'w')
errfile = open('err.log', 'w')

try:
  executableandargs = arguments
  executableandargs.insert(0, executable)
  child = subprocess.Popen(executableandargs, shell=False, stdout=outfile, stderr=errfile)

  while 1:
    result = child.poll()
    if result is not None:
        break
    #heartbeatfile.write('.')
    #flush_file(heartbeatfile)
    time.sleep(30)
    
except Exception as x:
  outfile.close()
  errfile.close()
  sys.stdout = sys.__stdout__
  sys.stderr = sys.__stderr__
  

outfile.close()
errfile.close()

## As a last step, we will create an archive of the outputsandbox files. This
## will speed up the post-staging process in many cases
try:
    filefilter
except:
    filefilter = None

from Ganga.Utility.files import multi_glob, recursive_copy

createPackedOutputSandbox(outputpatterns,filefilter,".")

sys.exit(result)
"""
        



def requiresAfsToken():
    return getLocalRoot().find('/afs') == 0


def getLocalRoot():
    from Ganga.Utility.files import fullpath
    import Ganga.Core.FileWorkspace
    return fullpath(Ganga.Core.FileWorkspace.gettop())



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ClassifyLocal.py,v 1.3 2008-11-12 11:56:43 karl Exp $
###############################################################################
# File: ClassifyLocal.py
# Author: K. Harrison
# Created: 070123

"""Module containing class dealing with preparation of jobs to run
   image-classification application on Local backend"""

__author__  = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__    = "7 November 2008"
__version__ = "1.3"

from Ganga.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from Ganga.GPIDev.Lib.File import  File
from Ganga.GPIDev.Lib.File import  FileBuffer
from Ganga.Utility import logging
from Ganga.Utility.files import fullpath

import os
import time

logger = logging.getLogger()
ptime = \
   "\"$(echo \"$(python -c \"import time; print '%.6f' % time.time()\")\")\""

class ClassifyLocal( IRuntimeHandler ):

   def prepare( self, app, appsubconfig, appmasterconfig, jobmasterconfig ):

      job = app.getJobObject()
      version = appsubconfig[ "version" ]
      libList = appsubconfig[ "libList" ]
      classifierDir = appsubconfig[ "classifierDir" ]
      outDir = appsubconfig[ "outDir" ]
      imageDir = appsubconfig[ "imageDir" ]
      elementList = imageDir.split( os.sep )
      imageList = appsubconfig[ "imageList" ]
      tagFile = appsubconfig[ "tagFile" ]

      lineList = []
      inbox = []
      outbox = []

      headList, headBox = self.head( job = job, version = version, \
         libList = libList, classifierDir = classifierDir )
      lineList.extend( headList )
      outbox.extend( headBox )

      bodyList, bodyBox = self.body( job = job, imageDir = imageDir, \
         imageList = imageList, tagFile = tagFile )
      lineList.extend( bodyList )
      inbox.extend( bodyBox )

      tailList, tailBox = self.tail( job = job )
      lineList.extend( tailList )
      outbox.extend( tailBox )

      jobScript = "\n".join( lineList )
      jobWrapper = FileBuffer( "Classify.sh", jobScript, executable = 1 )

      outbox.extend( job.outputsandbox )

      return StandardJobConfig\
         ( exe = jobWrapper, inputbox = inbox, outputbox = outbox )

   def head( self, job = None, version = "", libList = [], classifierDir = "" ):
      outbox = []
      lineList = \
         [
            "#!/bin/bash",
            "",
            "# Run script for image-classification application",
            "# Created by Ganga - %s" % ( time.strftime( "%c" ) ),
            "",
            "JOB_START_TIME=\\",
            "%s" % ptime,
            "HOSTNAME=$(hostname -f)",
            "echo \"Job running on ${HOSTNAME}\"",
            "echo \"Processor architecture: $(arch)\"",
            "echo \"Start time: $(date)\"",
            "",
            "WORKDIR=$(pwd)",
            "if [ -z ${PATH} ]; then",
            "   export PATH=\"\"",
            "fi",
            "if [ -z ${VO_CAMONT_SW_DIR} ]; then",
            "   VO_CAMONT_SW_DIR=\"\"",
            "else",
            "   VO_CAMONT_SW_DIR=${VO_CAMONT_SW_DIR}/%s" % version,
            "fi",
            "JAVA_INSTALLATION=${VO_CAMONT_SW_DIR}/java/$(arch)",
            "if [ -d ${JAVA_INSTALLATION} ]; then",
            "   JAVA_VERSION=`ls ${JAVA_INSTALLATION}`",
            "   JAVA_HOME=${JAVA_INSTALLATION}/${JAVA_VERSION}",
            "   JAVA=${JAVA_HOME}",
            "   JDK=${JAVA_HOME}/bin",
            "   PATH=${JDK}:${PATH}",
            "fi",
            "JAVALIB=\"\"",
         ]

      pathsep = ""
      for libPath in libList:
         libName = os.path.basename( libPath )
         lineList.extend\
            ( [
               "if [ -f %s ]; then" % libPath,
               "   JAVALIB=${JAVALIB}%s%s" % ( pathsep, libPath ),
               "elif [ -f ${WORKDIR}/%s ]; then" % libName,
               "   JAVALIB=${JAVALIB}%s${WORKDIR}/%s" % ( pathsep, libName ),
               # used to be ${VO_CAMONT_SW_DIR}/java/%s
               "elif [ -f ${VO_CAMONT_SW_DIR}/%s ]; then" % libName,
               "   JAVALIB=${JAVALIB}%s${VO_CAMONT_SW_DIR}/%s" \
                  % ( pathsep, libName ),
               "else",
               "   echo \"Unable to locate library \'%s\'\"" % libName,
               "   echo \"Searched in:\"",
               "   echo \"  %s\"" % os.path.dirname( libPath ),
               "   echo \"  ${WORKDIR}\"",
               "   echo \"  ${VO_CAMONT_SW_DIR}\"",
               "   exit",
               "fi",
            ] )
         if not pathsep:
            pathsep = os.pathsep

      lineList.extend\
         ( [
            "CLASSIFIER_HOME=\"\"",
            "if [ -d %s ]; then" % classifierDir,
            "   CLASSIFIER_HOME=%s" % classifierDir,
            "   IMENSE_LIB_DIR=%s/lib" % classifierDir,
            # used to be ${VO_CAMONT_SW_DIR}/classifier
            "elif [ -d ${VO_CAMONT_SW_DIR} ]; then",
            "   CLASSIFIER_HOME=${VO_CAMONT_SW_DIR}",
            "   IMENSE_LIB_DIR=${VO_CAMONT_SW_DIR}/lib",
            "else",
            "   echo \"Unable to locate classifier directory\"",
            "   exit",
            "fi",
            "if [ -z ${LD_LIBRARY_PATH} ]; then",
            "   export LD_LIBRARY_PATH=\"\"",
            "fi",
            "export LD_LIBRARY_PATH=${LD_LIBRARY_PATH}:${WORKDIR}",
            "export LD_LIBRARY_PATH=${LD_LIBRARY_PATH}:${IMENSE_LIB_DIR}"
            "",
         ] )
      return ( lineList, outbox )

   def body( self, job = None, imageDir = "", imageList = [], tagFile = "" ):

      lineList = []

      lineList.extend\
         ( [
         "DOWNLOAD_START_TIME=\\",
         "%s" % ptime,
         "N_DOWNLOAD_SUCCESS=0",
         "N_DOWNLOAD_FAILURE=0",
         ])

      if tagFile:
         xList, inbox, removeList = self.bodyTagFile( tagFile = tagFile )
      else:
         xList, inbox, removeList  = self.bodyUrlList( imageList = imageList )   
      lineList.extend( xList )

      lineList.extend\
         ( [
         "DOWNLOAD_END_TIME=\\",
         "%s" % ptime,
         "DOWNLOAD_SIZE=$(du -b -s ${WORKDIR}/images | cut -f 1)",
         ])

      lineList.extend\
         ( [
         "cd ${CLASSIFIER_HOME}",
         "COMMAND=\"$(which java) -Xmx900m -cp ${JAVALIB} "\
            +"imenseLib.ProcessImages -c ${WORKDIR}/images\"",
         "echo \"Performing image classification with command:\"",
         "echo \"   ${COMMAND}\"",
         "EXECUTION_START_TIME=\\",
         "%s" % ptime,
         "${COMMAND}",
         "EXECUTION_END_TIME=\\",
         "%s" % ptime,
         ] )

      lineList.extend( removeList )

      return ( lineList, inbox )

   def bodyTagFile( self, tagFile = "" ):

      lineList = []
      removeList = []
      inbox = [ File( name = tagFile ) ]

      lineList.extend\
         ( [
         "JOB_TAGFILE=\"%s\"" % os.path.basename( tagFile ),
         "TAGFILE_DIR=\"${WORKDIR}/tag_data\"",
         "rm -rf ${TAGFILE_DIR}",
         "mkdir -p ${TAGFILE_DIR}",
         "split -l 6 -a 7 ${JOB_TAGFILE} ${TAGFILE_DIR}/tagfile_",  
         "TAGFILE_LIST=$(ls ${TAGFILE_DIR})",
         "",
         "for TAGFILE in ${TAGFILE_LIST}; do",
         "  TAGFILE_PATH=${TAGFILE_DIR}/${TAGFILE}",
         "  IMAGE_URL=\"\"",
         "  while read TAG_LINE; do",
         "    TAG=${TAG_LINE%%:*}",
         "    if [ \"${TAG}\" == \"URLImage\" ]; then",
         "      IMAGE_URL=${TAG_LINE#\"URLImage: \"}",
         "      IMAGE_FILE=${IMAGE_URL##*/}",
         "      IMAGE_SUFFIX=${IMAGE_FILE##*\\.}",
         "      break",
         "    fi",
         "  done < ${TAGFILE_PATH}",
         "",
         "  if [ \"${IMAGE_URL}\" != \"\" ]; then"
         "    wget -q ${IMAGE_URL} -O ${IMAGE_FILE}",
         "    if [ -f ${IMAGE_FILE} ]; then",
         "      N_DOWNLOAD_SUCCESS=$((${N_DOWNLOAD_SUCCESS}+1))",
         "      IMAGE_MD5SUM=$(md5sum ${IMAGE_FILE})",
         "      IMAGE_MD5NAME=$(echo ${IMAGE_MD5SUM} | cut -d ' ' -f 1)",
         "      SUBDIR1=$(echo ${IMAGE_MD5NAME} | cut -c 1-3)",
         "      SUBDIR2=$(echo ${IMAGE_MD5NAME} | cut -c 4-6)",
         "      IMAGE_REMOTE_DIR=\"${WORKDIR}/images/${SUBDIR1}/${SUBDIR2}\"",
         "      IMAGE_PATH_NO_SUFFIX=\"${IMAGE_REMOTE_DIR}/${IMAGE_MD5NAME}\"",
         "      IMAGE_REMOTE_PATH=\"${IMAGE_PATH_NO_SUFFIX}.${IMAGE_SUFFIX}\"",
         "      TAGFILE_REMOTE_PATH=\"${IMAGE_PATH_NO_SUFFIX}.txt\"",
         "      mkdir -p ${IMAGE_REMOTE_DIR}/vseg",
         "      mv -f ${IMAGE_FILE} ${IMAGE_REMOTE_PATH}",
         "      mv -f ${TAGFILE_PATH} ${TAGFILE_REMOTE_PATH}",
         "    else",
         "     N_DOWNLOAD_FAILURE=$((${N_DOWNLOAD_FAILURE}+1))",
         "    fi",
         "  fi",
         "done",
         ""
         ])

      return ( lineList, inbox, removeList )

   def bodyUrlList( self, imageList = [] ):

      lineList = []
      removeList = []
      inbox = []

      for imageUrl, relativePath in imageList:

         if relativePath:
            imageFile = os.path.basename( relativePath )
            imageSubdir = os.path.join\
               ( "images", os.path.dirname( relativePath ) )
            imageRemoteDir = os.path.join( "${WORKDIR}", imageSubdir )
            imageRemotePath = os.path.join( imageRemoteDir, imageFile )
            vsegDir = os.path.join( imageRemoteDir, "vseg" )
            removeList.append( "rm %s" % imageRemotePath )
            lineList.extend\
               ( [
               "mkdir -p %s" % vsegDir,
               "wget -q -P %s '%s'" % ( imageRemoteDir, imageUrl ),
               "if [ -f %s ]; then" % imageRemotePath,
               "   N_DOWNLOAD_SUCCESS=$((${N_DOWNLOAD_SUCCESS}+1))",
               "else",
               "   N_DOWNLOAD_FAILURE=$((${N_DOWNLOAD_FAILURE}+1))",
               "fi",
#              "sleep 30s",
               ] )
         else:
            imageFile = os.path.basename( imageUrl )
            imageSuffix = os.path.splitext( imageFile )[ 1 ]  
            lineList.extend\
               ( [
               "wget -q %s -O %s" % ( imageUrl, imageFile ),
               "if [ -f %s ]; then" % imageFile,
               "   IMAGE_MD5SUM=$(md5sum %s)" % imageFile,
               "   IMAGE_MD5NAME=$(echo ${IMAGE_MD5SUM} | cut -d ' ' -f 1)",
               "   SUBDIR1=$(echo ${IMAGE_MD5NAME} | cut -c 1-3)",
               "   SUBDIR2=$(echo ${IMAGE_MD5NAME} | cut -c 4-6)",
               "   N_DOWNLOAD_SUCCESS=$((${N_DOWNLOAD_SUCCESS}+1))",
               ] )
            imageSubdir = os.path.join( "images", "${SUBDIR1}", "${SUBDIR2}" )
            imageRemoteDir = os.path.join( "${WORKDIR}", imageSubdir )
            imageRemotePath = os.path.join\
               ( imageRemoteDir, "${IMAGE_MD5NAME}" + imageSuffix )
            vsegDir = os.path.join( imageRemoteDir, "vseg" )
            lineList.extend\
               ( [
               "   mkdir -p %s" % vsegDir,
               "   mv -f %s %s" % ( imageFile, imageRemotePath ),
               "else",
               "   N_DOWNLOAD_FAILURE=$((${N_DOWNLOAD_FAILURE}+1))",
               "fi",
#              "sleep 30s",
               ] )

      return ( lineList, inbox, removeList )

   def tail( self, job = None ):
      lineList = \
         [
         "",
         "RESULTS_SIZE=$(du -b -s ${WORKDIR}/images | cut -f 1)",
         "cd ${WORKDIR}/images",
         "TAR_START_TIME=\\",
         "%s" % ptime,
         "tar -zcf ${WORKDIR}/images.tar.gz *",
         "TAR_END_TIME=\\",
         "%s" % ptime,
         "TARBALL_SIZE=$(du -b -s ${WORKDIR}/images.tar.gz | cut -f 1)",
         "",
         "RUN_DATA=${WORKDIR}/execute.dat",
         "rm -rf ${RUN_DATA}",
         "touch ${RUN_DATA}",
         "echo \"Hostname: ${HOSTNAME}\" >> ${RUN_DATA}",
         "echo \"Job_start: ${JOB_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Download_start: ${DOWNLOAD_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Download_end: ${DOWNLOAD_END_TIME}\" >> ${RUN_DATA}",
         "echo \"Download_successes: ${N_DOWNLOAD_SUCCESS}\" >> ${RUN_DATA}",
         "echo \"Download_failures: ${N_DOWNLOAD_FAILURE}\" >> ${RUN_DATA}",
         "echo \"Download_size: ${DOWNLOAD_SIZE}\" >> ${RUN_DATA}",
         "echo \"Execution_start: ${EXECUTION_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Execution_end: ${EXECUTION_END_TIME}\" >> ${RUN_DATA}",
         "echo \"Results_size: ${RESULTS_SIZE}\" >> ${RUN_DATA}",
         "echo \"Tar_start: ${TAR_START_TIME}\" >> ${RUN_DATA}",
         "echo \"Tar_end: ${TAR_END_TIME}\" >> ${RUN_DATA}",
         "echo \"Tarball_size: ${TARBALL_SIZE}\" >> ${RUN_DATA}",
         "JOB_END_TIME=\\",
         "%s" % ptime,
         "echo \"Job_end: ${JOB_END_TIME}\" >> ${RUN_DATA}",
         "echo \"End time: $(date)\"",
         ]
      outbox = [ "images.tar.gz", "execute.dat" ]
      return ( lineList, outbox )



#Clean

from GangaRobot.Framework.Action import IAction
from Ganga.Utility.logging import getLogger
from Ganga.Utility.Config import getConfig
import os, shutil
from os.path import join
from GangaRobot.Framework.exceptions import *

logger = getLogger()

class Publisher(IAction):
    """
    Publisher IAction implementation. This action will take the test results
    and pack them up in a .tgz archive in a published location. All the test
    files are renamed to have the location (site) of the tests included.
    """
    
    def execute(self, runid):
        logger.info("Starting publication")
        from Ganga.Utility.files import expandfilename
        config = getConfig('TestRobot')
        topdir=expandfilename(config['JobDir'])
        publishPath=expandfilename(config['PublishPath'])
        version = config['VersionNumber']
        site = config['Site']
        for i in range(len(config['TestPairs'])):
            jobdir = join(topdir, version+"_Job_"+str(i))
            import tarfile
            test=config['TestPairs'][i][0].replace('/','.')
            option=config['TestPairs'][i][2]
            fname = version+'_'+test+'_'+option+'-'+site+'.tgz'
            archive = tarfile.open(join(publishPath,fname),'w:gz')
            archive.posix=False
            for file in [f for f in os.listdir(join(jobdir,'output'))
                         if os.path.isfile(join(jobdir,'output',f))]:
                toname= file.replace(option,option+'-'+site)
                
                archive.add(join(jobdir,'output',file),join('output',toname))
            for file in [f for f in os.listdir(join(jobdir,'reports','latest'))
                         if os.path.isfile(join(jobdir,'reports','latest',f))]:
                toname = file.replace(option,option+'-'+site)
                archive.add(join(jobdir,'reports','latest',file),join(toname))
            archive.close()
